#!/usr/bin/env python
# Copyright (C) 2011-2012 CRS4.
#
# This file is part of Seal.
#
# Seal is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Seal is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Seal.  If not, see <http://www.gnu.org/licenses/>.


# Usage: test_bwa_aligner.py <genome reference - part common to all files>

from itertools import izip
import fileinput
import os
import StringIO
import sys
import unittest

from bl.lib.seq.aligner.bwa.bwa_aligner import BwaAligner
from bl.lib.seq.aligner.io.sam_formatter import SamFormatter
import bl.lib.seq.aligner.bwa.bwa_core as bwa
import testing_utilities as utils


class TestBwaAligner(unittest.TestCase):
  REFERENCE = os.path.join(os.path.dirname(__file__), 'fixtures/foobar.fa')
  BUILD_REFERENCE = True

  class SimpleVisitor(object):
    def __init__(self):
      self.sam = SamFormatter()
      self.output = StringIO.StringIO()

    def process(self, pair):
      for hit in pair:
        print >>self.output, self.sam.format(hit)

  def setUp(self):
    utils.build_ref_index()
    self.aligner = BwaAligner()
    self.aligner.reference = utils.reference
    self.aligner.hit_visitor = type(self).SimpleVisitor()

    self.pairs = []
    with open(utils.get_fixture_path("pairs.txt")) as f:
      for line in f:
        if not line.startswith("#"): # leave #-lines for comments
          self.pairs.append(line.rstrip("\r\n").split("\t"))

  def tearDown(self):
    utils.remove_ref_index()

  def test_load_clear_batch(self):
    for row in self.pairs:
      self.aligner.load_pair_record(row)
    self.assertEqual(len(self.pairs), self.aligner.get_batch_size())
    self.aligner.clear_batch()
    self.assertEqual(0, self.aligner.get_batch_size())

  def test_defaults(self):
    self.assertEqual("fastq-illumina", self.aligner.qformat)
    self.assertEqual(1000, self.aligner.max_isize)
    self.assertEqual(1, self.aligner.nthreads)
    self.assertEqual(0, self.aligner.trim_qual)

  def test_alignment(self):
    for row in self.pairs:
      self.aligner.load_pair_record(row)
    self.aligner.run_alignment()
    # TODO:  write a more useful test, but for that we'll need a complete test fixture
    self.assertTrue( len(self.aligner.hit_visitor.output.getvalue()) > 0 )

  def test_alignment_mmap(self):
    self.aligner.mmap_enabled = True
    # Generate the .sax and .rsax indices. They will be removed by tearDown.
    bwa.make_suffix_arrays_for_mmap(utils.reference)
    for row in self.pairs:
      self.aligner.load_pair_record(row)
    self.aligner.run_alignment()
    # TODO:  write a more useful test, but for that we'll need a complete test fixture
    self.assertTrue( len(self.aligner.hit_visitor.output.getvalue()) > 0 )

  def test_missing_mmap_index(self):
    self.aligner.mmap_enabled = True
    for row in self.pairs:
      self.aligner.load_pair_record(row)
    self.assertRaises(ValueError, self.aligner.run_alignment)


def suite():
  """Get a suite with all the tests from this module"""
  return unittest.TestLoader().loadTestsFromTestCase(TestBwaAligner)

if __name__ == '__main__':
  #if len(sys.argv) > 1:
  #  TestBwaAligner.REFERENCE = sys.argv[1]
  #  TestBwaAligner.BUILD_REFERENCE = False
  unittest.TextTestRunner(verbosity=2).run(suite())



# Copyright (C) 2011-2012 CRS4.
#
# This file is part of Seal.
#
# Seal is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Seal is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Seal.  If not, see <http://www.gnu.org/licenses/>.

import logging
logging.basicConfig(level=logging.DEBUG)
import time, os, glob, math
import itertools as it
import ctypes as ct

import Bio.SeqIO

import bl.lib.seq.aligner.bwa.bwa_core as bwa
from bl.lib.seq.aligner.bwa.bwa_aligner import BwaAligner
#import bl.lib.seq.generator as sg
from bl.lib.seq.aligner.io.sam_formatter import SamFormatter
import bl.lib.seq.aligner.bwa.constants as bwa_const
##---------------------------------##
# find or build the bwa executable
##---------------------------------##
if os.environ.has_key('SRC_DIR'): # provide a way to override the default location
  SRC_DIR = os.environ['SRC_DIR']
else:
  tree_root = os.path.abspath(os.path.join(os.path.dirname( __file__ ), '..', '..', '..', '..', '..', '..'))
  SRC_DIR = os.path.join(tree_root, 'bl', 'lib', 'seq', 'aligner', 'bwa', 'libbwa')
if not os.path.exists(SRC_DIR):
  raise ValueError("%r not found: redefine SRC_DIR in %s or in an environment variable" % (SRC_DIR, __name__))
BWA_EXE = os.path.join(SRC_DIR, 'bwa')
logging.debug("BWA_EXE: %r" % BWA_EXE)
if not os.path.exists(BWA_EXE):
  logging.warn("bwa executable not found; building source code")
  cmd = "make -C %s" % SRC_DIR
  ret = os.system(cmd)
  if ret:
    raise RuntimeError("%r failed -- could not make bwa executable" % cmd)
##---------------------------------##

##---------------------------------##
# Fixture helpers
##---------------------------------##
reference = os.path.join(os.path.abspath(os.path.dirname(__file__)), 'fixtures/foobar.fa')

def build_ref_index():
	if not os.path.exists('%s.bwt' % reference):
		os.system('%s index %s >/dev/null 2>&1' % (BWA_EXE, reference))

def remove_ref_index():
	os.system('rm -f %s.*' % reference)

##---------------------------------##

def clean_up_aux_files(fname):
  fs = glob.glob(fname + '.*')
  for f in fs:
    if os.path.splitext(f)[-1] != ".gz":
      os.unlink(f)

def create_seq_file(fname, seq, qtype="fastq-sanger"):
  fp = open(fname, 'w')
  fp_sanger = open("%s.sanger" % fname, 'w')
  if qtype == "fastq-illumina" or qtype == "fastq-solexa":
    hdr1, s, hdr2, q = seq.splitlines(False)
    # FIXME: duplicate code from bwa_core
    q = [ord(x)-bwa_core.Q_OFFSET[qtype] for x in q]
    if qtype == "fastq-solexa":
      q = [int(round(x+10*math.log10(1+10**(-x/10.)))) for x in q]
    q = "".join([chr(x+bwa_core.Q_OFFSET["fastq-sanger"]) for x in q])
    seq_sanger = "%s\n%s\n%s\n%s\n" % (hdr1, s, hdr2, q)
  else:
    seq_sanger = seq
  fp.write(seq)
  fp_sanger.write(seq_sanger)
  fp.close()
  fp_sanger.close()

def read_sam_file(fname):
  formatter = SamFormatter()
  fp = open(fname)
  res = []
  for l in fp:
    l = l.strip()
    if l[0] == "@":  # SAM header line
      continue
    r = formatter.parse(l)
    res.append(r)
  fp.close()
  return res

def build_index(refseq_fname):
  os.system('%s index %s' % (BWA_EXE, refseq_fname))

def run_bwa_samse(refseq_fname, hit_fname):
  hit_fname += ".sanger"
  os.system('%s aln %s %s > %s.aln' % (BWA_EXE, refseq_fname,
                                       hit_fname, hit_fname))
  os.system('%s samse %s %s.aln %s > %s.sam' % (BWA_EXE, refseq_fname,
                                                hit_fname, hit_fname,
                                                hit_fname))
  return read_sam_file(hit_fname + '.sam')

def run_bwa_sampe(refseq_fname, read_fname, mate_fname):
  read_fname += ".sanger"
  mate_fname += ".sanger"
  os.system('%s aln %s %s > %s.aln' % (BWA_EXE, refseq_fname,
                                       read_fname, read_fname))
  os.system('%s aln %s %s > %s.aln' % (BWA_EXE, refseq_fname,
                                       mate_fname, mate_fname))
  os.system('%s sampe %s %s.aln %s.aln %s %s > %s.sam' % (BWA_EXE, refseq_fname,
                                                          read_fname,
                                                          mate_fname,
                                                          read_fname,
                                                          mate_fname,
                                                          read_fname))
  return read_sam_file(read_fname + '.sam')

def run_bwa_py_sampe(refseq_fname, read_fname, mate_fname,
                     log_level=logging.INFO, pairing_batch_size=None,
                     seq_list_len=None, fastq_subfmt="fastq-illumina"):
  logger = logging.getLogger("PY")
  logger.setLevel(log_level)
  logger.info("RUNNING PYTHON VERSION")
  def debug_dump(seq, state):
    logger.debug("%s: name=%s" % (state, seq.get_name()))
    logger.debug("%s: qual=%s" % (state, seq.get_qual_string()))
    logger.debug("%s: strand=%d" % (state, seq.strand))
    logger.debug("%s: pos=%d" % (state, seq.pos))
    logger.debug("%s: mapQ=%d" % (state, seq.mapQ))

  read_flow = Bio.SeqIO.parse(open(read_fname), fastq_subfmt)
  mate_flow = Bio.SeqIO.parse(open(mate_fname), fastq_subfmt)
  pairs_flow = it.izip(read_flow, mate_flow)

  class ResultCollector(object):
    def __init__(self):
      self.result = []
    def process(self, pair):
      self.result.append(pair[0])
      self.result.append(pair[1])
  result = ResultCollector()

  while 1:
    start = time.time()
    pairs = list(it.islice(pairs_flow, 0, seq_list_len))
    if len(pairs) == 0:
      break
    # turn the biopython SeqRecords into simple tuples
    tuples = map(lambda t: (t[0].name, t[0].seq.tostring(), None, t[1].seq.tostring(), None), pairs[0:5])
    for t in tuples:
      print t
    logger.info('reading seqs %f sec' % (time.time() - start))

    start = time.time()
    aligner = BwaAligner()
    aligner.reference = refseq_fname
    aligner.hit_visitor = result
    for t in tuples[0:5]:
      aligner.load_pair_record(t)
    aligner.run_alignment()
    aligner.clear_batch()
    logger.info('alignment %f sec' % (time.time() - start))

  # map bwa mappings to dictionaries
  def bwam_to_hash(bwa_m):
    h = dict(
        name=bwa_m.name,
        aux=bwa_m.tags,
        seq=bwa_m.get_seq_5()
        )
    return h

  return map(bwam_to_hash, result.result)

def get_fixture_path(fixture_name):
	# get the path to this file's directory, then go into the fixtures directory
	# and finally attach the fixture_name
	return os.path.join(os.path.dirname(os.path.abspath( __file__ )), "fixtures", fixture_name)

base_key = {
    'A':0, 'a':0,
    'C':1, 'c':1,
    'G':2, 'g':2,
    'T':3, 't':3,
    'N':4, 'n':4
}
op_key = {
    'M':0,
    'I':1,
    'D':2,
    'S':3
}
key_base = 'ACGTN'
complement_key_base = 'TGCAN'

complement = {
  'A':'T',
  'C':'G',
  'G':'C',
  'T':'A',
  'N':'N'
}

def build_bwa_seq_t(hit):
  """Create a real bwa_seq_t from a dictionary containing its values"""
  struct = bwa.bwa_seq_t()
  struct.name = ct.c_char_p(hit['name'])
  struct.seq = ct.cast(ct.create_string_buffer(len(hit['seq'])), ct.POINTER(ct.c_uint8))
  struct.rseq = ct.cast(ct.create_string_buffer(len(hit['rseq'])), ct.POINTER(ct.c_uint8))
  struct.qual = ct.cast(ct.c_char_p(hit['qual']), ct.POINTER(ct.c_uint8))
  struct.len = hit['len']
  struct.strand = hit['strand']
  struct.type = hit['type']
  struct.dummy = hit['dummy']
  struct.extra_flag = hit['extra_flag']
  struct.n_mm = hit['n_mm']
  struct.n_gapo = hit['n_gapo']
  struct.n_gape = hit['n_gape']
  struct.mapQ = hit['mapQ']
  struct.score = hit['score']
  struct.clip_len = hit['clip_len']
  struct.n_aln = hit['n_aln']
  struct.aln = None
  struct.n_multi = hit['n_multi']
  struct.multi = None
  struct.sa = hit['sa']
  struct.pos = hit['pos']
  struct.c1 = hit['c1']
  struct.c2 = hit['c2']
  struct.seQ = hit['seQ']
  struct.n_cigar = hit['n_cigar']
  struct.cigar = ct.cast(ct.create_string_buffer( len(hit['cigar'])*ct.sizeof(ct.c_uint16) ), ct.POINTER(ct.c_uint16))
  struct.tid = hit['tid']
  struct.full_len = hit['full_len']
  struct.nm = hit['nm']
  struct.md = None

  for i in xrange(len(hit['seq'])):
    struct.seq[i] = base_key[ hit['seq'][i] ]
    struct.rseq[i] = base_key[ hit['rseq'][i] ]

  # create the CIGAR array
  for i in xrange(len(hit['cigar'])):
    size, op = hit['cigar'][i]
    encoded_v = (op << bwa_const.CIGAR_OP_SHIFT) | (size & bwa_const.CIGAR_LN_MASK)
    struct.cigar[i] = ct.c_uint16(encoded_v)

  return struct



# Copyright (C) 2011-2012 CRS4.
#
# This file is part of Seal.
#
# Seal is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Seal is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Seal.  If not, see <http://www.gnu.org/licenses/>.

# These are import path manipulations used by the bin scripts.

import os
import sys

# find the absolute path to the Seal directory
SealDir = os.path.realpath( os.path.join( os.path.dirname( os.path.realpath(__file__) ), "..") )

# now we set up sys.path and PYTHONPATH to include first the build directory, if it
# exists (we could be running from within the development copy), and second the SealDir
# (for a real installation).

# prepend SealDir to sys.path and PYTHONPATH
sys.path.insert(0, SealDir)
os.environ['PYTHONPATH'] = ":".join( filter(lambda x:x, [SealDir, os.environ.get('PYTHONPATH', None)]) )
BuildDir = os.path.join(SealDir, "build")
if os.path.exists(BuildDir):
	# since it exists, prepend BuildDir to sys.path and PYTHONPATH
	sys.path.insert(0,  BuildDir) # This is the build dir.
	os.environ['PYTHONPATH'] = ":".join([BuildDir, os.environ['PYTHONPATH'] ])



#!/usr/bin/env python

import fileinput
import os
import re
import sys

# expected output:
# CRESSIA_129:1:1:1003:1171#0	READ_1_SEQUENCE_1	READ_1_QUALITY_1	READ_2_SEQUENCE_1	READ_2_QUALITY_1
# CRESSIA_129:1:1:1010:1209#0	READ_1_SEQUENCE_3	READ_1_QUALITY_3	READ_2_SEQUENCE_3	READ_2_QUALITY_3
# etc etc...

read_pattern = re.compile("READ_(\d+)")
expected_reads = [ '1', '1', '2', '2' ]

sequence_pattern = re.compile("(?:SEQUENCE|QUALITY)_(\d+)")

if any( map(lambda s: s == 0, map(os.path.getsize, sys.argv[1:])) ):
  raise StandardError("we have at least one empty output file")

for line in fileinput.input(sys.argv[1:]):
  # ensure we have the two different reads
  reads = read_pattern.findall(line)
  if reads != expected_reads:
    raise StandardError("got unexpected reads pattern '%s' (expected %s)" % (reads, expected_reads))

  # ensure the line only references one sequence
  matches = sequence_pattern.findall(line)

  if len(matches) != 4:
    raise StandardError("expected 4 sequence ids by found %d" % len(matches))

  if any(matches[0] != matches[i] for i in range(len(matches))):
    raise StandardError("output record contains mismatched sequence fragments" % matches)



# Copyright (C) 2011-2012 CRS4.
#
# This file is part of Seal.
#
# Seal is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Seal is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Seal.  If not, see <http://www.gnu.org/licenses/>.

class jc_wrapper(object):
	"""
	Simple wrapper to provide a dict-list __setitem__ method to JobConf
	"""
	def __init__(self, jc):
		self.jc = jc
		self.cache = {}

	def __getitem__(self, k):
		if self.cache.has_key(k):
			return self.cache[k]
		else:
			return self.jc.get(k)

	def get(self, k):
		return self[k]

	def getInt(self, k):
		return int(self[k])

	def getFloat(self, k):
		return float(self[k])

	def getBoolean(self, k):
		return bool(self[k])

	def __setitem__(self, k, v):
		self.cache[k] = v

	def hasKey(self, k):
		return self.cache.has_key(k) or self.jc.hasKey(k)



# Copyright (C) 2011-2012 CRS4.
#
# This file is part of Seal.
#
# Seal is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Seal is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Seal.  If not, see <http://www.gnu.org/licenses/>.


######################################################
# Unit test for BwaMapping class
# To run:  SRC_DIR=<path to>/seqal/bl/lib/seq/aligner/bwa/libbwa  python test_bwa_mapping.py
#

import unittest
import ctypes as ct
import pickle
import sys
from itertools import izip

import bl.lib.seq.aligner.bwa.bwa_core as bwa
from bl.lib.seq.aligner.bwa.bwa_mapping import BwaMapping
from bl.lib.seq.aligner.sam_flags import *

import testing_utilities as utils

class FakeOpts(object):
	def __init__(self):
		# set the two options used by BwaMapping to their default values
		self.mode = 0x01 | 0x02 # BWA_MODE_GAPE | BWA_MODE_COMPREAD
		self.max_top2 = 30

class FakeReference(object):

	class Annotation(object):
		def __init__(self, name, offset, length):
			self.name = name
			self.offset = offset
			self.len = length

	def __init__(self):
		self.canned_results = {  # for bwa.get_seq_id
		  (1840870115L, 99L): (10L, 0L),
		  (1840869941L, 100L): (10L, 0L),
		  (1840869941L, 90L): (10L, 0L),
		  (1840870115L, 100L): (10L, 0L),
		  (406515435L, 56L): (1L, 0L),
		  (406515557L, 100L): (1L, 0L),
		  (406515557L, 35L): (1L, 0L),
		  (406515435L, 100L): (1L, 0L),
		  (867800532L, 100L): (3L, 0L),
		  (867800321L, 100L): (3L, 0L),
		  (867800321L, 100L): (3L, 0L),
		  (867800532L, 100L): (3L, 0L),
		  (414924637L, 100L): (1L, 0L),
		  (414924404L, 100L): (1L, 0L),
		  (414924404L, 80L): (1L, 0L),
		  (414924637L, 100L): (1L, 0L),
		  (556634939L, 91L): (2L, 0L),
		  (556634729L, 100L): (2L, 0L),
		  (556634729L, 99L): (2L, 0L),
		  (556634939L, 100L): (2L, 0L)
		}

		self.hg18_index_annotations = [
		  ('chr1', 0, 247249719L, 39L),
		  ('chr2', 247249719L, 242951149L, 25),
		  ('chr3', 490200868L, 199501827L, 10),
		  ('chr4', 689702695L, 191273063L, 14),
		  ('chr5', 880975758L, 180857866L, 7),
		  ('chr6', 1061833624L, 170899992L, 11),
		  ('chr7', 1232733616L, 158821424L, 12),
		  ('chr8', 1391555040L, 146274826L, 10),
		  ('chr9', 1537829866L, 140273252L, 38),
		  ('chr10', 1678103118L, 135374737L, 26),
		  ('chr11', 1813477855L, 134452384L, 9),
		  ('chr12', 1947930239L, 132349534L, 13),
		  ('chr13', 2080279773L, 114142980L, 6),
		  ('chr14', 2194422753L, 106368585L, 2),
		  ('chr15', 2300791338L, 100338915L, 11),
		  ('chr16', 2401130253L, 88827254L, 6),
		  ('chr17', 2489957507L, 78774742L, 10),
		  ('chr18', 2568732249L, 76117153L, 4),
		  ('chr19', 2644849402L, 63811651L, 5),
		  ('chr20', 2708661053L, 62435964L, 6),
		  ('chr21', 2771097017L, 46944323L, 11),
		  ('chr22', 2818041340L, 49691432L, 28),
		  ('chrX', 2867732772L, 154913754L, 16),
		  ('chrY', 3022646526L, 57772954L, 14),
		  ('chrM', 3080419480L, 16571L, 0)
		]

		# here we hold fake "anns" objects that respond to 'name'
		self.anns = [ FakeReference.Annotation(name, offset, length) for name, offset, length, unknown in self.hg18_index_annotations ]

	def get_seq_id(self, bns, hit_pos, hit_len):
		# We return hard-coded values for (position, length) combinations.
		# The returned values were calculated using the hg18 reference
		return self.canned_results[(hit_pos, hit_len)]



class TestBwaMapping(unittest.TestCase):

	def setUp(self):
		self.gap_opts = FakeOpts()
		self.reference = FakeReference()

		# To fake having a reference, we have to overwrite the bwa.get_seq_id function
		# called by BwaMapping so that it'll use our FakeReference instead of the real one.
		bwa.get_seq_id = lambda bns, pos, len: self.reference.get_seq_id(bns, pos, len)

		# now create the hits
		# With have a file with 10 serialized dictionaries containing values most fields of
		# the bwa_seq_t struct.  The items are in the order read1, mate1, read2, mate2, ...
		with open(utils.get_fixture_path("hit_wrapper_hits.pickle")) as f:
			hits_data = pickle.load(f) # load an array of dictionaries

		self.bwa_seqs = map(utils.build_bwa_seq_t, hits_data)
		# finally, create the BwaMapping objects for each bwa_seq_t
		self.hits = []
		for s1, s2 in izip( self.bwa_seqs[0::2], self.bwa_seqs[1::2] ):
			self.hits.append( BwaMapping(self.gap_opts, self.reference, s1, s2) )
			self.hits.append( BwaMapping(self.gap_opts, self.reference, s2, s1) )


	def test_is_paired(self):
		# all our hits are paired
		self.assertTrue(self.hits[0].is_paired())
		self.assertTrue(self.hits[1].is_paired())
		# modify a flag to unset any flags referencing a mate
		self.hits[0].flag = self.hits[0].flag & ~SAM_FPD & ~SAM_FPP & ~SAM_FMU & ~SAM_FMR & ~SAM_FR1 & ~SAM_FR2
		self.assertFalse(self.hits[0].is_paired())


	def test_is_properly_paired(self):
		self.assertTrue(self.hits[0].is_properly_paired())
		self.assertTrue(self.hits[1].is_properly_paired())
		self.hits[0].flag = self.hits[0].flag & ~SAM_FPP
		self.assertFalse(self.hits[0].is_properly_paired())


	def test_is_isnt_mapped(self):
		self.assertTrue(self.hits[0].is_mapped())
		self.assertTrue(self.hits[1].is_mapped())
		# fake flag
		self.hits[0].flag = self.hits[0].flag | SAM_FSU
		self.assertFalse(self.hits[0].is_mapped())
		self.assertTrue(self.hits[0].is_unmapped())

	def test_is_mate_mapped_unmapped(self):
		self.assertTrue(self.hits[0].is_mate_mapped())
		self.assertTrue(self.hits[1].is_mate_mapped())
		# fake flag
		self.hits[0].flag = self.hits[0].flag | SAM_FMU
		self.assertFalse(self.hits[0].is_mate_mapped())
		self.assertTrue(self.hits[0].is_mate_unmapped())

	def test_is_on_reverse(self):
		# indices [0, 3, 4, 6, 8] are on reverse strand
		hit = self.hits[0]; self.assertTrue( hit.is_on_reverse())
		hit = self.hits[3]; self.assertTrue( hit.is_on_reverse())
		hit = self.hits[4]; self.assertTrue( hit.is_on_reverse())
		hit = self.hits[6]; self.assertTrue( hit.is_on_reverse())
		hit = self.hits[1]; self.assertFalse( hit.is_on_reverse())
		hit = self.hits[2]; self.assertFalse( hit.is_on_reverse())


	def test_is_mate_on_reverse(self):
		# indices [0, 3, 4, 6, 8] are on reverse strand
		hit = self.hits[1]; self.assertTrue( hit.is_mate_on_reverse())
		hit = self.hits[2]; self.assertTrue( hit.is_mate_on_reverse())
		hit = self.hits[5]; self.assertTrue( hit.is_mate_on_reverse())
		hit = self.hits[7]; self.assertTrue( hit.is_mate_on_reverse())
		hit = self.hits[0]; self.assertFalse( hit.is_mate_on_reverse())
		hit = self.hits[3]; self.assertFalse( hit.is_mate_on_reverse())
		hit = self.hits[4]; self.assertFalse( hit.is_mate_on_reverse())


	def test_is_read1(self):
		self.assertTrue( self.hits[0].is_read1())
		self.assertFalse( self.hits[1].is_read1())
		self.assertTrue( self.hits[2].is_read1())
		self.assertFalse( self.hits[3].is_read1())

	def test_is_read2(self):
		self.assertFalse( self.hits[0].is_read2())
		self.assertTrue( self.hits[1].is_read2())
		self.assertFalse( self.hits[2].is_read2())
		self.assertTrue( self.hits[3].is_read2())

	def test_is_secondary_align(self):
		# we don't have any secondary alignments
		self.assertTrue( all(map(lambda hit: not hit.is_secondary_align(), self.hits)))
		self.hits[0].flag = self.hits[0].flag | SAM_FSC
		self.assertTrue(self.hits[0].is_secondary_align())

	def test_tag_value(self):
		# we should have these flags on hits[0]:
		# XC:i:99 XT:A:U  NM:i:0  SM:i:37 AM:i:37 X0:i:1  X1:i:0  XM:i:0  XO:i:0  XG:i:0  MD:Z:99
		hit = self.hits[0]
		self.assertEqual(99, hit.tag_value("XC"))
		self.assertEqual('U', hit.tag_value("XT"))
		self.assertEqual(0, hit.tag_value("NM"))
		self.assertEqual(37, hit.tag_value("SM"))
		self.assertEqual(37, hit.tag_value("AM"))
		self.assertEqual(1, hit.tag_value("X0"))
		self.assertEqual(0, hit.tag_value("X1"))
		self.assertEqual(0, hit.tag_value("XM"))
		self.assertEqual(0, hit.tag_value("XO"))
		self.assertEqual(0, hit.tag_value("XG"))
		# we can't check the MD tag because our struct doesn't include the md data
		# self.assertEqual('99', hit.tag_value("MD"))
		# what happens with a missing tag?
		self.assertTrue(hit.tag_value("lallala") is None) # assertIsNone isn't in the version of python we're using

	def test_flag_string(self):
		self.assertEqual("pPr1", self.hits[0].flag_string())
		self.assertEqual("pPR2", self.hits[1].flag_string())
		# make up some flags
		hit = self.hits[0]
		hit.flag = SAM_FPD | SAM_FSU | SAM_FR1
		self.assertEqual("pu1", hit.flag_string())
		hit.flag = SAM_FPD | SAM_FMU | SAM_FR1
		self.assertEqual("pU1", hit.flag_string())

	def __format_cigar(self, cigar_array):
		return "".join(['%d%s' % t for t in cigar_array]) or "*"

	def test_cigar(self):
		for i, cigar in (0, "1S99M"), (1, "90M10S"), (2, "56M44S"), (3, "65S35M"), (4, "100M"), (5, "100M"):
			self.assertEqual(cigar, self.__format_cigar(self.hits[i].get_cigar()), "%s != %s -- %s" % (cigar, self.__format_cigar(self.hits[i].get_cigar()), str(self.hits[i])))

	def test_isize(self):
		for i, isize in (0, -273), (1, 273), (2, 157), (3, -157), (4, -311), (5, 311):
			self.assertEqual(isize, self.hits[i].isize)

	def test_pos(self):
		for i, pos in (0, 27392261), (1, 27392087), (2, 159265717), (3, 159265839), (4, 178097838), (5, 178097627):
			self.assertEqual(pos, self.hits[i].pos)

	def test_mpos(self):
		for i, pos in (0, 27392087), (1, 27392261), (2, 159265839), (3, 159265717), (4, 178097627), (5, 178097838):
			self.assertEqual(pos, self.hits[i].mpos)

	def test_tid(self):
		for i, tid in (0, "chr11"), (1, "chr11"), (2, "chr2"), (3, "chr2"), (4, "chr4"), (5, "chr4"):
			self.assertEqual(tid, self.hits[i].tid)

	def test_mtid(self):
		for i, mtid in (0, "="), (1, "="), (2, "="), (3, "="), (4, "="), (5, "="):
			self.assertEqual(mtid, self.hits[i].mtid)


	def test_name(self):
		names = ['CRESSIA:1:1:10000:101092#0/1',
		'CRESSIA:1:1:10000:101092#0/2',
		'CRESSIA:1:1:10000:104835#0/1',
		'CRESSIA:1:1:10000:104835#0/2',
		'CRESSIA:1:1:10000:105809#0/1',
		'CRESSIA:1:1:10000:105809#0/2',
		'CRESSIA:1:1:10000:112583#0/1',
		'CRESSIA:1:1:10000:112583#0/2',
		'CRESSIA:1:1:10000:72482#0/1',
		'CRESSIA:1:1:10000:72482#0/2']

		for i in xrange(len(names)):
			self.assertEqual(names[i], self.hits[i].get_name())

	def test_qual(self):
		## we know that our version gives mapq values that are different from the command line version
		## BWA.  We'll test to ensure that the values in our BwaAlignments correspond to the values
		## in the bwa structure.  Arguably, to write a pure unit test for BwaMapping, we should be
		## only considering the incoming bwa_seq_t values for all the other tests as well...
		#for i in xrange(len(self.hits)):
		#	self.assertEqual(self.bwa_seqs[i].mapQ, self.hits[i].qual)
		for i in xrange(len(self.hits)):
			self.assertEqual(60, self.hits[i].qual)


	def test_get_seq_5(self):
		# indices [0, 3, 4, 6, 8] are on reverse strand
		expected_answers = [
				(0, "CRESSIA:1:1:10000:101092#0/1", "TTAAAAAAGCTTCCCCAAGACCATTCAATAAGCAGTGTAACAGAAAACACCAGAGCCCATGACCTTAAGTGCAAACGATGAGTCTTCTTAAGTATTATTT"),
				(1, "CRESSIA:1:1:10000:101092#0/2", "TACTAGCCTTTCCCAAAGCAGGTTTGGAAAAGTTTTTGGCTTGGGAATAACAGTAACAGTAGCAGCAACAAAAACAGCAGGAGCAGCAGCTAAGATGCAG"),
				(4, "CRESSIA:1:1:10000:105809#0/1", "TTCCATTTTTGCCAGCAGACTAACCCCAGACTCTAAATCCCAGTAGCAATGTCTTCAGCTTGTGGGTGGTACTGGCTTCTTGCTGTTAGTATTCTGTGGG"),
				(5, "CRESSIA:1:1:10000:105809#0/2", "AAACGAAATCTTCATGCCAAAACTAACACTTGTCAGTTGGACCTCTCTATAGTCTGTAACGTATTTTATTTCCTCTCCCTTTAGTGAAGTTCATCTGACA")
		]

		for idx, name, seq in expected_answers:
			self.assertEqual(name, self.hits[idx].get_name())
			self.assertEqual(seq, self.hits[idx].get_seq_5())

	def test_qualseq(self):
		expected_answers = [
				(0, "CRESSIA:1:1:10000:101092#0/1", "##EEEEEE=EF:DDDFGGGGEEEBEFDGEFBFG=FEEDBGGGGGGGEGDGGGGGFFGFGGGGDGGGFGGEGEGGGGGGEGGGGGGGGEGFGGGGFGGGGG"),
				(1, "CRESSIA:1:1:10000:101092#0/2", "GGGGGGDBGGGFGGGGGDEGGGGGGEGGGGFGGGGGEGGFGGDGA<BAAAEEEEDBDDDBGEGADDDDD:GGGGGEGDD?BA::B??AA###########"),
				(4, "CRESSIA:1:1:10000:105809#0/1", ":=DD=ECECAAA?>ADBD@DEDCEEEDGGEGDFEGFECGGGFGFFGGGGGGBFGGEGGGGGFGGGGGEGGGGGGGGGGGGGDGGGGGGGGFGGGFGGGGG"),
				(5, "CRESSIA:1:1:10000:105809#0/2", "GGGGGGGGGGGGGGGGGGGGGGGGFGGGGGGGGGFGFGGGGGGGGGGGDGFFFFFGFGFGFEDFFGGFGGDGGGGGGGFCGDGGF?FFFEGGGFGGGGG?")
		]
		for idx, name, qseq in expected_answers:
			self.assertEqual(name, self.hits[idx].get_name())
			self.assertEqual(qseq, ''.join([ chr(q+33) for q in self.hits[idx].get_base_qualities() ]) )
			self.assertEqual(qseq, self.hits[idx].get_ascii_base_qual())

	def test_tags(self):
		clip_tags = map(lambda h: filter(lambda tple: tple[0] == "XC", h.each_tag()), self.hits)
		for i in xrange(len(self.hits)):
			xc_tag = filter(lambda tpl: tpl[0] == "XC", self.hits[i].each_tag())
			if self.bwa_seqs[i].full_len > self.bwa_seqs[i].clip_len:
				self.assertTrue(len(xc_tag) > 0) # ensure we have an XC tag
				xc_value = xc_tag[0][2]
				self.assertEqual(self.bwa_seqs[i].clip_len, xc_value)

	def test_get_seq_len(self):
		for h in self.hits:
			self.assertEqual(100, h.get_seq_len())

	"""
	0 XC: 99 on reverse? True
	1 XC: 90 on reverse? False
	2 XC: 56 on reverse? False
	3 XC: 35 on reverse? True
	4 XC: None on reverse? True
	5 XC: None on reverse? False
	6 XC: None on reverse? True
	7 XC: 80 on reverse? False
	8 XC: 91 on reverse? True
	9 XC: 99 on reverse? False
	"""
	def test_get_untrimmed_pos_unclipped_forward(self):
		# hits[5] has not been clipped and it's on the forward strand
		self.assertEqual(self.hits[5].pos, self.hits[5].get_untrimmed_left_pos())

	def test_get_untrimmed_pos_clipped_forward(self):
		# hits[2] has been clipped to length 90, and it's not on the reverse strand
		self.assertEqual(self.hits[2].pos, self.hits[2].get_untrimmed_left_pos())

	def test_get_untrimmed_pos_unclipped_reverse(self):
		# hits[4] has not been clipped and it's on the reverse strand
		self.assertEqual(self.hits[4].pos, self.hits[4].get_untrimmed_left_pos())

	def test_get_untrimmed_pos_clipped_reverse(self):
		# hits[3] has been clipped to 35 and it's on the reverse strand
		self.assertEqual(self.hits[3].pos - 65, self.hits[3].get_untrimmed_left_pos())

	def test_base_qualities_out_of_range(self):
		bwa_seq = self.bwa_seqs[0]
		for i in xrange(0, bwa_seq.len):
			bwa_seq.qual[i] = 1 # insert a bad encoded quality value
		mapping = BwaMapping(self.gap_opts, self.reference, bwa_seq, self.bwa_seqs[1])
		self.assertRaises(ValueError, mapping.get_base_qualities)

	def test_xa(self):
		bwa_seq = self.bwa_seqs[0]
		mate = self.bwa_seqs[1]
		# add an alternative hit
		bwa_seq.n_multi = 1
		# copy the alignment from another bwa_seq
		multi = bwa.bwt_multi1_t()
		alternative = self.bwa_seqs[3]
		alternative_hit = self.hits[3] # the BwaMapping for the same hit
		multi.pos = alternative.pos
		multi.n_cigar = alternative.n_cigar
		multi.cigar = alternative.cigar
		multi.n_gapo = alternative.n_gapo
		multi.n_gape = alternative.n_gape
		multi.n_mm = alternative.n_mm
		multi.strand = alternative.strand
		multi.score = alternative.score
		bwa_seq.multi = ct.pointer(multi)

		# create a new BwaMapping object
		mapping = BwaMapping(self.gap_opts, self.reference, bwa_seq, mate)

		# get the XA tag and test its value
		xa_tag = mapping.tag_value("XA")
		self.assertFalse(xa_tag is None)
		elements = xa_tag.rstrip(";").split(";")
		self.assertEqual(1, len(elements))
		parts = elements[0].split(",")
		self.assertEqual(4, len(parts))
		self.assertEqual(alternative_hit.tid, parts[0])
		self.assertEqual(('-' if alternative_hit.is_on_reverse() else '+') + str(alternative_hit.pos), parts[1])
		self.assertEqual(alternative_hit.get_cigar_str(), parts[2])
		self.assertEqual(str(alternative.n_mm), parts[3])


def suite():
	"""Get a suite with all the tests from this module"""
	return unittest.TestLoader().loadTestsFromTestCase(TestBwaMapping)

if __name__ == '__main__':
	unittest.TextTestRunner(verbosity=2).run(suite())



# Copyright (C) 2011-2012 CRS4.
#
# This file is part of Seal.
#
# Seal is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Seal is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Seal.  If not, see <http://www.gnu.org/licenses/>.

import random
import re

from bl.lib.seq.aligner.mapping import SimpleMapping

def erase_read1(pair):
	pair[0] = None
	p1 = pair[1]
	p1.set_mate_mapped(False)
	p1.mtid = None
	p1.m_ref_id = None
	p1.mpos = 0
	return pair

def erase_read2(pair):
	pair[1] = None
	p0 = pair[0]
	p0.set_mate_mapped(False)
	p0.mtid = None
	p0.m_ref_id = None
	p0.mpos = 0
	return pair

def pair1():
	p = (SimpleMapping(), SimpleMapping())
	p[0].set_name("p1:read/1") ; p[1].set_name("p1:read/2")
	p[0].tid = "chr1" ; p[1].tid = "chr1"
	p[0].ref_id = 0 ; p[1].ref_id = 0
	p[0].pos = 12345  ; p[1].pos = p[0].pos + 150
	p[0].set_read1(True) ; p[1].set_read2(True)
	p[0].qual = 50 ; p[1].qual = 30
	return p

def pair2():
	# reversed reads such that r2 is at a lower position than r1
	p = (SimpleMapping(), SimpleMapping())
	p[0].set_name("p1:read/1") ; p[1].set_name("p1:read/2")
	p[0].tid = "chr1" ; p[1].tid = "chr1"
	p[0].ref_id = 0 ; p[1].ref_id = 0
	p[0].pos = 12345 + 150 ; p[1].pos = p[0].pos - 150
	p[0].set_read1(True) ; p[1].set_read2(True)
	p[0].qual = 50 ; p[1].qual = 30
	return p


def make_key(mapping):
	from bl.mr.seq.seqal.mapper import MarkDuplicatesEmitter
	# This function must be the same as MarkDuplicatesEmitter::get_hit_key
	return MarkDuplicatesEmitter.get_hit_key(mapping)



# Copyright (C) 2011-2012 CRS4.
#
# This file is part of Seal.
#
# Seal is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Seal is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Seal.  If not, see <http://www.gnu.org/licenses/>.

import logging
logging.basicConfig(level=logging.DEBUG)
import unittest, time, math, itertools as it
import os
import re

import Bio.SeqIO
from Bio.Alphabet import single_letter_alphabet
from Bio.Seq import Seq
from Bio.SeqRecord import SeqRecord
import bl.lib.seq.generator as sg

import bl.lib.seq.aligner.bwa as bwa
from bl.lib.seq.generator import reverse_complement
import testing_utilities as u
import testing_data as d


class bwa_base_tc(unittest.TestCase):

  def setUp(self):
    self.refseq_fn, self.read_fn, self.mate_fn = 'full_match', 'read', 'mate'

  def tearDown(self):
    for fn in [self.refseq_fn, self.read_fn, self.mate_fn]:
      u.clean_up_aux_files(fn)

  def __do_pair_align_and_comparison(self, log_level=logging.WARNING,
                                     pairing_batch_size=None,
                                     seq_list_len=None,
                                     fastq_subfmt="fastq-sanger"):
    u.build_index(self.refseq_fn)
    #--
    start = time.time()
    bwa_res = u.run_bwa_sampe(self.refseq_fn, self.read_fn, self.mate_fn)
    print 'BWA_SAMPE: %f secs' % (time.time() - start)
    #--
    start = time.time()
    py_res =  u.run_bwa_py_sampe(self.refseq_fn, self.read_fn, self.mate_fn,
                                 log_level, pairing_batch_size, seq_list_len,
                                 fastq_subfmt)
    print 'PYBWA_SAMPE: %f secs' % (time.time() - start)
    #--
    self.assertEqual(len(bwa_res), len(py_res))
    for o, n in it.izip(bwa_res, py_res):
      for k in o.keys():
        if k == 'name':
          nn = n[k].rsplit("/", 1)[0]  # we are dropping the '/[12]' extension
          self.assertEqual(o[k], nn)
        elif k == 'aux':
          self.assertEqual(len(o[k]), len(n[k]))
          for x,y in it.izip(o[k], n[k]):
            self.assertEqual(x, y)
        else:
          self.assertEqual(o[k],n[k])

  def hand_picked(self):
    u.create_seq_file(self.refseq_fn, d.refseq, qtype="fasta")
    u.create_seq_file(self.read_fn, d.hit_1_seq, qtype="fastq-sanger")
    u.create_seq_file(self.mate_fn, d.hit_2_seq, qtype="fastq-sanger")
    self.__do_pair_align_and_comparison()

  def hand_picked_real_1(self):
    u.create_seq_file(self.refseq_fn, d.chr22_chunk, qtype="fasta")
    for t, f in (("read", self.read_fn), ("mate", self.mate_fn)):
      seq = getattr(d, "real_case_1_%s" % t)
      u.create_seq_file(f, seq, qtype="fastq-illumina")
    self.__do_pair_align_and_comparison(fastq_subfmt="fastq-illumina")

  def hand_picked_real_2(self):
    u.create_seq_file(self.refseq_fn, d.mm_chr17_chunk, qtype="fasta")
    for t, f in (("read", self.read_fn), ("mate", self.mate_fn)):
      seq = getattr(d, "real_case_2_%s" % t)
      u.create_seq_file(f, seq, qtype="fastq-sanger")
    self.__do_pair_align_and_comparison(fastq_subfmt="fastq-sanger")

  def auto_generated(self):
    u.create_seq_file(self.refseq_fn, d.refseq, qtype="fasta")
    n_seq, seq_len, span = 1000, 75, 500
    refseq = Bio.SeqIO.parse(open(self.refseq_fn), 'fasta').next()
    pairs = u.generate_pairs(refseq.seq, n_seq, seq_len, span=span)
    for i, fn in enumerate([self.read_fn, self.mate_fn]):
      fp = open(fn, 'w')
      Bio.SeqIO.write([x[i] for x in pairs], fp, 'fastq-illumina')
      fp.close()
      fp = open("%s.sanger" % fn, 'w')
      Bio.SeqIO.write([x[i] for x in pairs], fp, 'fastq-sanger')
      fp.close()
    self.__do_pair_align_and_comparison(fastq_subfmt="fastq-illumina")

  def batch_pairing(self):
    u.create_seq_file(self.refseq_fn, d.refseq, qtype="fasta")
    n_seq, seq_len, span = 1000, 75, 500
    refseq = Bio.SeqIO.parse(open(self.refseq_fn), 'fasta').next()
    pairs = u.generate_pairs(refseq.seq, n_seq, seq_len, span=span)
    for i, fn in enumerate([self.read_fn, self.mate_fn]):
      fp = open(fn, 'w')
      Bio.SeqIO.write([x[i] for x in pairs], fp, 'fastq-illumina')
      fp.close()
      fp = open("%s.sanger" % fn, 'w')
      Bio.SeqIO.write([x[i] for x in pairs], fp, 'fastq-sanger')
      fp.close()
    pairing_batch_size = 101
    self.__do_pair_align_and_comparison(pairing_batch_size=pairing_batch_size,
                                        fastq_subfmt="fastq-illumina")

  def multi_ref(self):
    refseq = d.refseq + d.chr22_chunk
    u.create_seq_file(self.refseq_fn, refseq, qtype="fasta")
    seq_len, span = 75, 500
    f = open(self.refseq_fn)
    raw_seqs = [r.seq.tostring() for r in Bio.SeqIO.parse(f, 'fasta')]
    f.close()
    raw_refseq = "".join(raw_seqs)
    starts = [
      0,                             # start of full refseq
      len(raw_seqs[0]) - seq_len/2,  # bridges the two seqs
      len(raw_refseq) - span         # end of full refseq
      ]
    pairs = u.generate_localized_pairs(raw_refseq, starts, seq_len, span=span)
    for i, fn in enumerate([self.read_fn, self.mate_fn]):
      fp = open(fn, 'w')
      Bio.SeqIO.write([x[i] for x in pairs], fp, 'fastq-illumina')
      fp.close()
      fp = open("%s.sanger" % fn, 'w')
      Bio.SeqIO.write([x[i] for x in pairs], fp, 'fastq-sanger')
      fp.close()
    self.__do_pair_align_and_comparison(fastq_subfmt="fastq-illumina")

  def bioseq_to_bwa_seq(self):
    nseq = 100
    for fmt in "fastq-sanger", "fastq-illumina", "fastq-solexa":
      qkey = "solexa_quality" if fmt == "fastq-solexa" else "phred_quality"
      for unknown in False, True:
        g = u.random_reads_generator(nseq, fmt=fmt, unknown=unknown)
        for i, (seq, q) in enumerate(g):
          name = "foo-%d" % i
          seq_str = "".join(seq)
          bioseq = SeqRecord(Seq(seq_str, single_letter_alphabet),
                             id=name, name=name, description=name)
          bioseq.letter_annotations[qkey] = q
          n, m = len(bioseq), len(bioseq.name)
          bwseq = bwa.alloc_seq(1, n, m)[0]
          bwa.bioseq_to_bwa_seq(bioseq, bwseq, n, m, fmt)
          self.assertEqual(bioseq.name, bwseq.get_name())
          self.assertEqual(bioseq.seq.data[::-1], bwseq.get_seq())
          self.assertEqual(reverse_complement(bioseq.seq.data),
                           bwseq.get_rseq())
          # check that quality has been converted to sanger
          if fmt == "fastq-solexa":
            exp_q = [int(round(x+10*math.log10(1+10**(-x/10.)))) for x in q]
          else:
            exp_q = q
          exp_qstr = "".join(chr(x+sg.Q_OFFSET["fastq-sanger"]) for x in exp_q)
          self.assertEqual(bwseq.get_qual(), exp_qstr)

  def build_bws_array(self):
    # FIXME: plenty of UGLY code
    nseq = 100
    for fmt in "fastq-sanger", "fastq-illumina", "fastq-solexa":
      qkey = "solexa_quality" if fmt == "fastq-solexa" else "phred_quality"
      for unknown in False, True:
        g = u.random_reads_generator(nseq, fmt=fmt, unknown=unknown, pe=True)
        bioseq_pairs, qseq_pairs = [], []
        for i, read_pair in enumerate(g):
          base_name = "foo-%d" % i
          names = ["%s/%d" % (base_name, j) for j in 1, 2]
          seq_strings = ["".join(read_pair[j][0]) for j in 0, 1]
          q_strings = ["".join(chr(x+sg.Q_OFFSET[fmt]) for x in read_pair[j][1])
                       for j in 0, 1]
          bioseq_p = [SeqRecord(Seq(seq_strings[j], single_letter_alphabet),
                                id=names[j],
                                name=names[j],
                                description=names[j]) for j in 0, 1]
          for j in 0, 1:
            bioseq_p[j].letter_annotations[qkey] = read_pair[j][1]
          bioseq_pairs.append(bioseq_p)
          qseq_pairs.append([base_name])
          for j in 0, 1:
            for l in seq_strings, q_strings:
              qseq_pairs[-1].append(l[j])
        n = len(bioseq_pairs[0][0])
        assert len(qseq_pairs[0][1]) == n
        for src in "bioseq", "qseq":
          seq_pairs = bioseq_pairs if src == "bioseq" else qseq_pairs
          bwsa = bwa.build_bws_array(seq_pairs, qtype=fmt, src=src)
          for j in 0, 1:
            self.assertTrue(type(bwsa[j]) is bwa.bwa_seq_p_t)
            for i in xrange(nseq):
              bwseq = bwsa[j][i]
              self.assertEqual(bwseq.len, n)
              if src == "bioseq":
                exp_name = bioseq_pairs[i][0].name
              else:
                exp_name = "%s/%d" % (qseq_pairs[i][0], (j+1))
              self.assertEqual(len(bwseq.get_name()), len(exp_name))

  # This test simply checks that some sequences have been trimmed.
  # It makes no attempt to verify the correctness of the trimming,
  # since we rely directly on BWA for the implementation of the
  # trimming algorithm.
  def build_bws_array_with_trim_qual(self):
    pairs = self.load_pairs_fixture("pairs.txt")
    bws_array = bwa.build_bws_array(pairs, qtype="fastq-illumina", src="qseq", trim_qual=15)
    sequences = [ bws_array[read][i] for i in xrange(5) for read in 0,1 ]
    clipped = map(lambda seq: seq.full_len > seq.clip_len, sequences)
    self.assertTrue(any(clipped))


  # This test checks that no sequences have been trimmed.
  def build_bws_array_with_trim_qual_zero(self):
    pairs = self.load_pairs_fixture("pairs.txt")
    bws_array = bwa.build_bws_array(pairs, qtype="fastq-illumina", src="qseq", trim_qual=0)
    for read in 0,1:
      for i in xrange(len(pairs)):
        seq = bws_array[read][i]
        self.assertEqual(seq.full_len, seq.clip_len, "unexpected sequence trimming")

  def load_pairs_fixture(self, fixture_name):
    pairs = []
    with open( u.get_fixture_path(fixture_name), 'r' ) as f:
      for line in f:
        if not re.match("\s*#", line): # if line doesn't start with #
          pairs.append(line.rstrip("\n\r").split("\t"))
    return pairs

def suite():
  suite = unittest.TestSuite()
  suite.addTest(bwa_base_tc('hand_picked'))
  suite.addTest(bwa_base_tc('hand_picked_real_1'))
  suite.addTest(bwa_base_tc('hand_picked_real_2'))
  suite.addTest(bwa_base_tc('auto_generated'))
  suite.addTest(bwa_base_tc('multi_ref'))
  suite.addTest(bwa_base_tc('batch_pairing'))
  suite.addTest(bwa_base_tc('bioseq_to_bwa_seq'))
  suite.addTest(bwa_base_tc('build_bws_array'))
  suite.addTest(bwa_base_tc('build_bws_array_with_trim_qual'))
  suite.addTest(bwa_base_tc('build_bws_array_with_trim_qual_zero'))
  return suite


if __name__ == '__main__':
  runner = unittest.TextTestRunner(verbosity=2)
  runner.run((suite()))



# Copyright (C) 2011-2012 CRS4.
#
# This file is part of Seal.
#
# Seal is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Seal is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Seal.  If not, see <http://www.gnu.org/licenses/>.

class HitProcessorChainLink(object):
	def __init__(self, next_link = None):
		self.next_link = next_link

	def set_next(self, link):
		self.next_link = link
		return link

	def process(self, pair):
		"""
		Identity action.  Passes to the next link, if any.
		"""
		if self.next_link:
			self.next_link.process(pair)




# Copyright (C) 2011-2012 CRS4.
#
# This file is part of Seal.
#
# Seal is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option)
# any later version.
#
# Seal is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
# for more details.
#
# You should have received a copy of the GNU General Public License along
# with Seal.  If not, see <http://www.gnu.org/licenses/>.


import copy
import os
import sys
import subprocess
import glob

import bl
SealJarName = "seal.jar"

def __is_exe(fpath):
	return os.path.exists(fpath) and os.access(fpath, os.X_OK)

def num_nodes():
	hproc = subprocess.Popen([hadoop, "job", "-list-active-trackers"], stdout=subprocess.PIPE)
	stdout, stderr = hproc.communicate()
	if hproc.returncode == 0:
		return stdout.count("\n") # trackers are returned one per line
	else:
		raise RuntimeError("Error running hadoop job -list-active-trackers")

def hdfs_path_exists(path):
	retcode = subprocess.call([hadoop, 'dfs', '-stat', path], stdout=open('/dev/null', 'w'), stderr=subprocess.STDOUT)
	return retcode == 0

def run_hadoop_cmd_e(cmd, properties=None, args_list=[]):
	retcode = run_hadoop_cmd(cmd, properties, args_list)
	if retcode != 0:
		raise RuntimeError("Error running Hadoop command")

def run_hadoop_cmd(cmd, properties=None, args_list=[]):
	args = [hadoop, cmd]
	if properties:
		args += __construct_property_args(properties)
	args += map(str, args_list) # only string arguments are allowed
	return subprocess.call(args)

def dfs(*args):
	return run_hadoop_cmd_e("dfs", args_list=args)

def run_hadoop_jar(jar_name, properties=None, args_list=[]):
	if os.path.exists(jar_name) and os.access(jar_name, os.R_OK):
		args = [hadoop, 'jar', jar_name]
		if properties:
			args += __construct_property_args(properties)
		args += map(str, args_list)
		retcode = subprocess.call(args)
		if retcode != 0:
			raise RuntimeError("Error running hadoop jar")
	else:
		raise ValueError("Can't read jar file %s" % jar_name)

def __construct_property_args(prop_dict):
	return sum(map(lambda pair: ["-D", "%s=%s" % pair], prop_dict.iteritems()), []) # sum flattens the list

def run_class_e(class_name, additional_cp=None, properties=None, args_list=[]):
	retcode = run_class(class_name, additional_cp, properties, args_list)
	if retcode != 0:
		raise RuntimeError("Error running Hadoop class")

def run_pipes(executable, input_path, output_path, properties=None, args_list=[]):
	args = [hadoop, "pipes"]
	properties = properties.copy() if properties else {}
	properties['hadoop.pipes.executable'] = executable

	args.extend( __construct_property_args(properties) )
	args.extend(["-input", input_path, "-output", output_path])
	args.extend(args_list)
	return subprocess.call(args)


def run_class(class_name, additional_cp=None, properties=None, args_list=[]):
	"""
	Run a class that needs the Hadoop jars in its class path
	"""
	args = [hadoop, class_name]
	if additional_cp:
		env = copy.copy(os.environ)
		if type(additional_cp) == str: # wrap a single class path in a list
			additional_cp = [additional_cp]
		env['HADOOP_CLASSPATH'] = ":".join(additional_cp)
	else:
		env = os.environ
	if properties:
		args.extend( __construct_property_args(properties) )
	args.extend(args_list)
	return subprocess.call(args)

def find_jar(jar_name, root_path=None):
	root = root_path or os.getcwd()
	paths = (root, os.path.join(root, "build"), "/usr/share/java")
	for path in [ os.path.join(path, jar_name) for path in paths ]:
		if os.path.exists(path):
			return path
	return None

def find_seal_jar():
  return os.path.join(bl.__path__[0], SealJarName)

#################################################################################
# module initialization
#################################################################################

hadoop = None

if os.environ.has_key("HADOOP_HOME") and \
	__is_exe(os.path.join(os.environ["HADOOP_HOME"], "bin", "hadoop")):
	hadoop = os.path.join(os.environ["HADOOP_HOME"], "bin", "hadoop")
else:
	# search the PATH for hadoop
	for path in os.environ["PATH"].split(os.pathsep):
		hpath = os.path.join(path, 'hadoop')
		if __is_exe(hpath):
			hadoop = hpath
			break
	if hadoop is None:
		raise ImportError("Couldn't find hadoop executable.  Please set HADOOP_HOME or add the hadoop executable to your PATH")
	hadoop = os.path.abspath(hadoop)




# Copyright (C) 2011-2012 CRS4.
#
# This file is part of Seal.
#
# Seal is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Seal is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Seal.  If not, see <http://www.gnu.org/licenses/>.

import unittest
import ctypes as ct
import os
import re
import struct
import sys

import bl.lib.seq.aligner.bwa.bwa_core as bwa
import testing_utilities as tu

class TestBwaCore(unittest.TestCase):
	def setUp(self):
		self.toy_data = [
			[ "name1", "CAAAT", "latte", "CAGAT", "acqua" ],
			[ "name2", "TTAGG", "beefy", "CTTGG", "caffe" ],
		]
		self.toy_bwsa = bwa.build_bws_array(self.toy_data, qtype="fastq-illumina")

	def tearDown(self):
		for seq_array in self.toy_bwsa:
			bwa.free_seq(len(self.toy_data), seq_array)


	def test_bwa_init_sequences_no_trim(self):
		row_type = (ct.c_char_p * 3)
		pointers = (row_type*len(self.toy_data))()
		q_offset = bwa.Q_OFFSET["fastq-illumina"]
		for i, row in enumerate(self.toy_data):
			for j in xrange(3):
				pointers[i][j] = row[j]
		array = bwa.init_sequences( ct.cast(pointers, bwa.p_char_p), len(self.toy_data), q_offset, 0)

		resulting_qoffset = q_offset - bwa.Q_OFFSET["fastq-sanger"]
		for i, row in enumerate(self.toy_data):
			self.assertEqual(-1, array[i].tid)
			self.assertEqual(len(row[1]), array[i].len)
			self.assertEqual(len(row[1]), array[i].full_len)
			self.assertEqual(len(row[1]), array[i].clip_len)
			self.assertEqual(row[0], array[i].get_name())
			self.assertEqual(row[1][::-1], array[i].get_seq())
			self.assertEqual(row[2], ''.join( [chr(array[i].qual[k]+resulting_qoffset) for k in xrange(array[i].full_len)] ))

		bwa.free_seq(len(self.toy_data), array)

	def test_init_sequences_with_trimming(self):
		data = []
		data.append(("seq1", "NNNNNNNNNNNNNNNNNNNNNNNNNNNNNNNNNNNNNNNNNNNNNNNNNNNNNNNNNNNNNNNNNNNNNNNNNNNNNNNNNNNNNNNNNNNNNNNNNNNN",
				"####################################################################################################"))
		row_type = (ct.c_char_p * 3)
		pointers = (row_type*len(data))()
		q_offset = bwa.Q_OFFSET["fastq-sanger"]
		for i, row in enumerate(data):
			for j in xrange(3):
				pointers[i][j] = row[j]
		array = bwa.init_sequences( ct.cast(pointers, bwa.p_char_p), len(data), q_offset, 15)
		self.assertEqual(35, array[0].clip_len)
		bwa.free_seq(len(data), array)


	def test_build_bws_array_simple(self):
		for i in xrange(len(self.toy_data)):
			self.assertEqual(self.toy_data[i][0], self.toy_bwsa[0][i].get_name())
			self.assertEqual(self.toy_data[i][0], self.toy_bwsa[1][i].get_name())

	def test_trimming_min_length(self):
		pair = ( # (name, r1, q1, r2, q2)
			("seq_name",
			 "NNNNNNNNNNNNACAGCTTTTGAGNGGGGCNCAGCCTCTTCCTTGTGGTGTTGCAGGACGGCAGGGAGTCAGTTGGGTTTCNACTCTTTAAGGACAGTGA",
			 "####################################################################################################",
			"TTTTTATTTCTAGTGTTTATATTGATGAACAGAAATATACTGACATATTAACTTTTTTCATATAAATTTTTTCAAATTTTTGGTTAAGGCTTTTTCTGTC",
			"GGGGGF@EGGD;EEEGG??EDEEDAE==E?EEE??>B6C:=CDCBGFGFFFGFGGF=EF5<?=)=9B6?BAA-A:;;>;5*2*57AA>AA>AA>CA@CED"),
		)
		# The first sequence should have a clipped length of 35 at q=15.
		# 35 is the minimum trimmed sequence length.
		bwsa = bwa.build_bws_array(pair, qtype="fastq-sanger", trim_qual=15)
		self.assertEqual(100, bwsa[0][0].full_len)
		self.assertEqual(35, bwsa[0][0].clip_len)
		self.assertEqual(35, bwsa[0][0].len)
		self.assertEqual(100, bwsa[1][0].clip_len) # this one isn't clipped at q=15
		for seq_array in bwsa:
			bwa.free_seq(len(pair), seq_array)

	def test_build_bws_array_with_trim_qual(self):
		pair = ( # read, qual, trimmed length at qual=15
			("pair1",
			 "ATNCACCTGCCTTGGCCTCGCAAAGTGCTGGGATTACAGGCATGANNNNNNNNNNNNNNNNNNCNGTNNNNNNNNNNNNNTNCANATNNNNNNNNNNNTG",
			 "==#@7::;;?FF?FFGGGFFGGGGE=CDEAE=AE?BBC??@###########################################################",
			 "TTATAAGCACTGCACATAACTTTCTCCCTAATCTTTACAACAATCNNNTNNNCNNANNNNNCNCAGGNNNNCNNGNNNNNGNCANTGNNNNNNNNNNNGT",
			 "EFDFFEEC?BEEAEE=A?CCE??EEAEEEEEECECEDEAC5@;7?#######################################################"
			),
		)
		lengths = (42, 46)
		bwsa = bwa.build_bws_array(pair, qtype="fastq-sanger", trim_qual=15)
		self.assertEqual(lengths[0], bwsa[0][0].clip_len)
		self.assertEqual(lengths[1], bwsa[1][0].clip_len)
		for seq_array in bwsa:
			bwa.free_seq(len(pair), seq_array)

	def test_build_bws_array_with_trim_qual_zero(self):
		pair = ( # read, qual, trimmed length at qual=15
			("pair1",
			 "ATNCACCTGCCTTGGCCTCGCAAAGTGCTGGGATTACAGGCATGANNNNNNNNNNNNNNNNNNCNGTNNNNNNNNNNNNNTNCANATNNNNNNNNNNNTG",
			 "==#@7::;;?FF?FFGGGFFGGGGE=CDEAE=AE?BBC??@###########################################################",
			 "TTATAAGCACTGCACATAACTTTCTCCCTAATCTTTACAACAATCNNNTNNNCNNANNNNNCNCAGGNNNNCNNGNNNNNGNCANTGNNNNNNNNNNNGT",
			 "EFDFFEEC?BEEAEE=A?CCE??EEAEEEEEECECEDEAC5@;7?#######################################################"
			),
		)
		bwsa = bwa.build_bws_array(pair, qtype="fastq-sanger", trim_qual=0)
		self.assertEqual(100, bwsa[0][0].clip_len)
		self.assertEqual(100, bwsa[1][0].clip_len)
		for seq_array in bwsa:
			bwa.free_seq(len(pair), seq_array)

	def test_build_bws_array_with_illumina_q(self):
		pair = (
			("pair1",
			"GCGTTAGTTTGTGGTGAAAGAAGCAAAATATTATGAATTTTGAAAATCTAGAGTCAATCTTCAAATTTCTTTTTATTCATTAATATCTCTGGCCTTTGTT",
			"TcLacaU_b_a\\aa`L\\a]X`dddYM^Z\\W]]W]WWbbYTZLZLZRI[VYQVWUU^b\\ccTc\\c^cYcYcZ^_ULUVUX[bY`^BBBBBBBBBBBBBBBB",
			"ATCAATCTAGGAGTTTGTATATACCACAAGAACAATGCTTGTCAAATCACTCACTTCTTTTCCCCTCTCTGGAGCCAGGCTTCGAGAAGTTGTATCCAGA",
			"B_a`aaTT__Q\\UQV^Z```\\VRKYQHUVIRLQQTTK]`^Z]]Y]_Y_Yb`da`^`aaac]eKdddfdefefecffffdaffffffefdedcffdeffff"),
		)
		lengths = (85,100)
		bwsa = bwa.build_bws_array(pair, qtype="fastq-illumina", trim_qual=15)
		self.assertEqual(lengths[0], bwsa[0][0].clip_len)
		self.assertEqual(lengths[1], bwsa[1][0].clip_len)
		for seq_array in bwsa:
			bwa.free_seq(len(pair), seq_array)

	def test_build_bws_array_without_qual_str(self):
		# make tuples substituting None for the quality strings in toy_data
		toy = map(lambda t: (t[0], t[1], None, t[3], None), self.toy_data)
		bwsa = bwa.build_bws_array(toy)
		for i in xrange(len(self.toy_data)):
			# verify name, sequence (which is reversed)
			self.assertEqual(self.toy_data[i][0], bwsa[0][i].get_name())
			self.assertEqual(self.toy_data[i][0], bwsa[1][i].get_name())
			self.assertEqual(self.toy_data[i][1], bwsa[0][i].get_seq()[::-1])
			self.assertEqual(self.toy_data[i][3], bwsa[1][i].get_seq()[::-1])
			# ensure that the returned quality string is empty
			self.assertEqual('', bwsa[0][i].get_qual_string())
			self.assertEqual('', bwsa[1][i].get_qual_string())
		for seq_array in bwsa:
			bwa.free_seq(len(toy), seq_array)

	def test_bwa_seq_t_get_seq(self):
		# sequence in bwa_seq_t is reversed at this stage
		self.assertEqual(self.toy_data[0][1], self.toy_bwsa[0][0].get_seq()[::-1])
		self.assertEqual(self.toy_data[0][3], self.toy_bwsa[1][0].get_seq()[::-1])
		self.assertEqual(self.toy_data[1][1], self.toy_bwsa[0][1].get_seq()[::-1])
		self.assertEqual(self.toy_data[1][3], self.toy_bwsa[1][1].get_seq()[::-1])

	def test_bwa_seq_t_get_rseq(self):
		# sequence in bwa_seq_t is reversed at this stage, but the reverse
		# complement sequence is itself reversed.  We can therefore compute the
		# current reverse complement by complementing each base in the original
		# DNA fragment.
		self.assertEqual(''.join(tu.complement[c] for c in self.toy_data[0][1]), self.toy_bwsa[0][0].get_rseq())
		self.assertEqual(''.join(tu.complement[c] for c in self.toy_data[0][3]), self.toy_bwsa[1][0].get_rseq())
		self.assertEqual(''.join(tu.complement[c] for c in self.toy_data[1][1]), self.toy_bwsa[0][1].get_rseq())
		self.assertEqual(''.join(tu.complement[c] for c in self.toy_data[1][3]), self.toy_bwsa[1][1].get_rseq())

	def test_bwa_seq_t_get_pos_5_raises(self):
		# on new, unaligned sequences it should raise
		self.assertRaises(ValueError, self.toy_bwsa[0][0].get_pos_5)

	def test_make_mmap_sa(self):
		try:
			tu.build_ref_index()
			bwa.make_suffix_arrays_for_mmap(tu.reference)

			sax_name = "%s.sax" % tu.reference
			rsax_name = "%s.rsax" % tu.reference
			self.assertTrue(os.path.exists(sax_name))
			self.assertTrue(os.path.exists(rsax_name))
			# check magic
			with open(sax_name) as f:
				s = f.read(4)
				self.assertEqual(4, len(s))
				self.assertEqual(bwa.mmap_magic().value, struct.unpack("=I", s)[0])

			with open(rsax_name) as f:
				s = f.read(4)
				self.assertEqual(4, len(s))
				self.assertEqual(bwa.mmap_magic().value, struct.unpack("=I", s)[0])
		finally:
			tu.remove_ref_index()

def suite():
	"""Get a suite with all the tests from this module"""
	return unittest.TestLoader().loadTestsFromTestCase(TestBwaCore)

if __name__ == '__main__':
	unittest.TextTestRunner(verbosity=2).run(suite())



# Copyright (C) 2011-2012 CRS4.
#
# This file is part of Seal.
#
# Seal is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Seal is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Seal.  If not, see <http://www.gnu.org/licenses/>.

import unittest
import copy

from bl.lib.seq.aligner.sam_mapping import SAMMapping

class TestSamMapping(unittest.TestCase):
	def setUp(self):
		self.sam1 = "HWI-ST200R_251:5:1208:19924:124635#GCCAAT\t83\t20\t6181935\t60\t5S96M\t=\t6181919\t-112\tAAGTGGAAGATTTGGGAATCTGAGTGGATTTGGTAACAGTAGAGGGGTGGATCTGGCTTGGAAAACAATCGAGGTACCAATATAGGTGGTAGATGAATTTT\t?<?AADADBFBF<EHIGHGGGEAF3AF<CHGGDG9?GHFFACDHH)?@AHEHHIIIIE>A=A:?);B27@;@?>,;;C(5::>>>@5:()4>@@@######\tXC:i:96\tXT:A:U\tNM:i:1\tSM:Z:raw_merged-1.2.3.4.5.6.7.8.bam\tAM:i:37\tX0:i:1\tX1:i:0\tXM:i:1\tXO:i:0\tXG:i:0\tMD:Z:13G82\tRG:Z:raw_merged-1.2.3.4.5.6.7.8.bam\tOQ:Z:######@@@>4)(:5@>>>::5(C;;,>?@;@72B;)?:A=A>EIIIIHHEHA@?)HHDCAFFHG?9GDGGHC<FA3FAEGGGHGIHE<FBFBDADAA?<?"
		self.sam2 = "HWI-ST200R_251:5:1208:19924:124635#GCCAAT\t163\t20\t6181919\t60\t101M\t=\t6181935\t112\tCTGAGCACACCAAAATTCATCTACCACCTATATTGGTACCTCGATTGTTTTCCAAGCCAGATCCACACCTCTACTGTTACCAAATCCACTCAGATTCCCAA\t@@@FFFDDFHG??;EEH>HHGIGHEGCGEGGIGJG31?DDBBD>FGG@HG??DFBBADFAGII3@EH;;CEHECBB7?>CE.;...5>ACDDA:C:;>:>?\tXT:A:U\tNM:i:2\tSM:Z:raw_merged-1.2.3.4.5.6.7.8.bam\tAM:i:37\tX0:i:1\tX1:i:0\tXM:i:2\tXO:i:0\tXG:i:0\tMD:Z:29G36C34\tRG:Z:raw_merged-1.2.3.4.5.6.7.8.bam\tOQ:Z:@@@FFFDDFHG??;EEH>HHGIGHEGCGEGGIGJG31?DDBBD>FGG@HG??DFBBADFAGII3@EH;;CEHECBB7?>CE.;...5>ACDDA:C:;>:>?"
		self.maps = map(SAMMapping, (self.sam1, self.sam2) )

	def test_ref_id(self):
		self.assertEqual(20, self.maps[0].ref_id)

	def test_tid(self):
		self.assertEqual("20", self.maps[0].tid)

	def test_pos(self):
		self.assertEqual(6181935, self.maps[0].pos)

	def test_name(self):
		self.assertEqual("HWI-ST200R_251:5:1208:19924:124635#GCCAAT", self.maps[0].get_name())

	def test_flag(self):
		self.assertEqual(83, self.maps[0].flag)

	def test_mpos(self):
		self.assertEqual(6181919, self.maps[0].mpos)

	def test_mtid(self):
		self.assertEqual("=", self.maps[0].mtid)

	def test_qual(self):
		self.assertEqual(60, self.maps[0].qual)

	def test_isize(self):
		self.assertEqual(-112, self.maps[0].isize)

	def test_is_paired(self):
		self.assertTrue(self.maps[0].is_paired())

	def test_is_properly_paired(self):
		self.assertTrue(self.maps[0].is_properly_paired())

	def test_is_mapped(self):
		self.assertTrue(self.maps[0].is_mapped())

	def test_is_unmapped(self):
		self.assertFalse(self.maps[0].is_unmapped())

	def test_is_mate_mapped(self):
		self.assertTrue(self.maps[0].is_mate_mapped())

	def test_is_mate_unmapped(self):
		self.assertFalse(self.maps[0].is_mate_unmapped())

	def test_is_on_reverse(self):
		self.assertTrue(self.maps[0].is_on_reverse())
		self.assertFalse(self.maps[1].is_on_reverse())

	def test_is_mate_on_reverse(self):
		self.assertFalse(self.maps[0].is_mate_on_reverse())
		self.assertTrue(self.maps[1].is_mate_on_reverse())

	def test_is_read1(self):
		self.assertTrue(self.maps[0].is_read1())
		self.assertFalse(self.maps[1].is_read1())

	def test_is_read2(self):
		self.assertFalse(self.maps[0].is_read2())
		self.assertTrue(self.maps[1].is_read2())

	def test_is_secondary_align(self):
		self.assertFalse(self.maps[0].is_secondary_align())

	def test_get_untrimmed_pos_unclipped_forward(self):
		# hit has not been clipped and it's on the forward strand
		self.assertEqual(self.maps[1].pos, self.maps[1].get_untrimmed_pos())

	def test_get_untrimmed_pos_clipped_reverse(self):
		# hit has been clipped to 96 and it's on the reverse strand
		self.assertEqual(self.maps[0].pos - 5, self.maps[0].get_untrimmed_pos())

	def test_get_seq_len(self):
		self.assertEqual(101, self.maps[0].get_seq_len())

	def test_difficult_tag(self):
		self.assertEqual("######@@@>4)(:5@>>>::5(C;;,>?@;@72B;)?:A=A>EIIIIHHEHA@?)HHDCAFFHG?9GDGGHC<FA3FAEGGGHGIHE<FBFBDADAA?<?", self.maps[0].tag_value("OQ"))

def suite():
	"""Get a suite with all the tests from this module"""
	return unittest.TestLoader().loadTestsFromTestCase(TestSamMapping)

if __name__ == '__main__':
	unittest.TextTestRunner(verbosity=2).run(suite())



# Copyright (C) 2011-2012 CRS4.
#
# This file is part of Seal.
#
# Seal is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Seal is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Seal.  If not, see <http://www.gnu.org/licenses/>.

import bwa_core as bwa
from bwa_mapping import BwaMapping

from bl.lib.tools.event_monitor import QuietMonitor
from bl.lib import TIME_TICK

class BWAIterator(object):

  def __init__(self, root, gopt, popt, max_isize, pairing_batch_size,
               visitor=None, mmap_enabled=False):
    if visitor is None:
      self.visitor = QuietMonitor()
    else:
      self.visitor = visitor
    self.root = root
    self.gopt = gopt
    self.popt = popt
    self.max_isize = max_isize
    self.pairing_batch_size = pairing_batch_size
    self.mmap_enabled = mmap_enabled
    # reference structures, loaded on demand
    self.__bwts = None # the reference index
    # the reference itself
    self.__pacseq = None
    self.__bnsp = None

  def reference_loaded(self):
    # we use a single variable (__bwts) to determine whether
    # the ref has been loaded.
    return self.__bwts is not None

  def load_reference(self):
    """
    Load the reference at self.root.
    Remember to deallocate it when unload_reference() once finished!"""
    if self.reference_loaded():
      raise RuntimeError("A reference is already loaded!")
    try:
      with self.visitor.time_block("restore_index"):
        self.__bwts = bwa.restore_index(self.root, self.mmap_enabled)

      with self.visitor.time_block("restore_reference"):
        self.__bnsp, self.__pacseq = bwa.restore_reference(self.root,
                                                           self.mmap_enabled)
    except:
      self.unload_reference()
      raise

  def unload_reference(self):
    """Free an allocated reference, if any.
    Sets __bwts, __pacseq, and __bnsp to None.
    """
    if self.__bwts:
      for j in 0, 1:
        bwa.bwt_destroy(self.__bwts[j], self.mmap_enabled)
      self.__bwts = None

    if self.__pacseq:
      del self.__pacseq
      self.__pacseq = None

    if self.__bnsp:
      bwa.bns_destroy(self.__bnsp, self.mmap_enabled)
      self.__bnsp = None

  def __analyze(self, bwsa, seq_pairs_read):
    ##################################################
    # This method performs the actual analysis work.
    # BWA writes its results directly into the bwsa structures.
    ##################################################
    if not self.reference_loaded():
      self.load_reference()

    ii = bwa.isize_info_t()
    last_ii = bwa.isize_info_t()
    last_ii.avg = -1.0

    self.visitor.start("cal_sa_reg_gap")
    for i in 0, 1:
      bwa.cal_sa_reg_gap_mt(self.__bwts, seq_pairs_read, bwsa[i], self.gopt)
    self.visitor.stop("cal_sa_reg_gap")

    self.visitor.start("cal_pac_pos_pe")
    cnt_chg = bwa.cal_pac_pos_pe(self.__bwts, seq_pairs_read, bwsa, ii,
                                 self.popt, self.gopt, last_ii)
    self.visitor.stop("cal_pac_pos_pe")

    if ii.avg > self.max_isize:
      self.visitor.warn("skipping S-W, isize is too big (%.3f)" % ii.avg)
    else:
      self.visitor.start("paired_sw")
      if 0 < self.pairing_batch_size < seq_pairs_read:
        for offset in xrange(0, seq_pairs_read, self.pairing_batch_size):
          nseq = min(self.pairing_batch_size, seq_pairs_read - offset)
          self.visitor.start("paired_sw_batch")
          bwa.paired_sw(self.__bnsp, self.__pacseq, nseq, bwsa, self.popt, ii, offset)
          self.visitor.stop_batch("paired_sw_batch", offset, nseq)
        else:
          bwa.paired_sw(self.__bnsp, self.__pacseq, seq_pairs_read, bwsa, self.popt, ii)
      self.visitor.stop("paired_sw")

    self.visitor.start("refine_gapped")
    for i in 0, 1:
      bwa.refine_gapped(self.__bnsp, seq_pairs_read, bwsa[i], self.__pacseq)
    self.visitor.stop("refine_gapped")


  def analyze(self, bwsa, seq_pairs_read):
    """
    Align the seq_pairs_read sequence pairs in bwsa and iterate
    through the matches one read at a time.  For each read
    in each pair in bwsa, this method will yield
    pairs of BwaMapping objects (mread1, mread2) and (mread2, mread1).
    """
    self.__analyze(bwsa, seq_pairs_read)
    for i in xrange(seq_pairs_read):
      for j in 0, 1:
        yield BwaMapping(self.gopt[0], self.__bnsp[0], bwsa[j][i], bwsa[j^1][i])


  def analyze_pairs(self, bwsa, seq_pairs_read):
    """
    Align the seq_pairs_read sequence pairs in bwsa and iterate
    through the matches two reads at a time.  For each pair
    in bwsa this method will yield a
    pairs of BwaMapping objects (mread1, mread2) and (mread2, mread1).

    To analyze the match objects use bwa.analyze_hit
    """
    self.__analyze(bwsa, seq_pairs_read)
    for i in xrange(seq_pairs_read):
      # XXX LP: if we leave in this check our filtered map counts will be wrong.
      # since it doesn't seem to make much of a difference in speed I'll leave it in.
      #if bwsa[0][i].type == bwa.BWA_TYPE_NO_MATCH or bwsa[1][i].type == bwa.BWA_TYPE_NO_MATCH:
        yield BwaMapping(self.gopt[0], self.__bnsp[0], bwsa[0][i], bwsa[1][i]), BwaMapping(self.gopt[0], self.__bnsp[0], bwsa[1][i], bwsa[0][i])




# Copyright (C) 2011-2012 CRS4.
#
# This file is part of Seal.
#
# Seal is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Seal is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Seal.  If not, see <http://www.gnu.org/licenses/>.

import os, tempfile, shutil, subprocess as sp

EXE = os.path.join(
  os.environ["HOME"],
  "svn/ac-dc/bl/apps/seqal/trunk/bin/find_fq_format.py"
  )

# 3rd seq is sanger, the others are undecidable
SANGER_INPUT = r"""@foo
AAAAAAAAAAAAAAA
+foo
;<=>?@ABCDEFGHI
@foo
AAAAAAAAAAAAAAAA
+foo
;<=>?@ABCDEFGHI
@foo
AAAAAAAAAAAAAAAA
+foo
:;<=>?@ABCDEFGHI
@foo
AAAAAAAAAAAAAAA
+foo
;<=>?@ABCDEFGHI
"""

# 3rd seq is illumina, the others are undecidable
ILLUMINA_INPUT = r"""@foo
AAAAAAAAAAAAAAA
+foo
;<=>?@ABCDEFGHI
@foo
AAAAAAAAAAAAAAAA
+foo
;<=>?@ABCDEFGHI
@foo
AAAAAAAAAAAAAAAA
+foo
;<=>?@ABCDEFGHIJ
@foo
AAAAAAAAAAAAAAA
+foo
;<=>?@ABCDEFGHI
"""


def make_file(tempd, basename, content):
  base_sanger_fn = basename
  sanger_fn = os.path.join(tempd, base_sanger_fn)
  f = open(sanger_fn, 'w')
  f.write(content)
  f.close()
  return sanger_fn


tempd = tempfile.mkdtemp(prefix="try_find_fq_format")

sanger_fn = make_file(tempd, "sanger.fq", SANGER_INPUT)
output = sp.Popen([EXE, sanger_fn], stdout=sp.PIPE).communicate()[0]
print "expected = sanger, result =", output
output = sp.Popen([EXE, sanger_fn, '-n 2'], stdout=sp.PIPE).communicate()[0]
print "expected = unknown, result =", output

illumina_fn = make_file(tempd, "sanger.fq", ILLUMINA_INPUT)
output = sp.Popen([EXE, illumina_fn], stdout=sp.PIPE).communicate()[0]
print "expected = illumina, result =", output
output = sp.Popen([EXE, illumina_fn, '-n 2'], stdout=sp.PIPE).communicate()[0]
print "expected = unknown, result =", output

shutil.rmtree(tempd)



# Copyright (C) 2011-2012 CRS4.
#
# This file is part of Seal.
#
# Seal is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Seal is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Seal.  If not, see <http://www.gnu.org/licenses/>.


import io
import unittest

from bl.lib.tools.seal_config_file import SealConfigFile, FormatError

class TestSealConfigFile(unittest.TestCase):

	CfgDefaultWithTitle = \
"""[DEFAULT]
key1:value1
key2:value2
"""

	CfgSection1 = \
"""[DEFAULT]
key1:value1
key2:value2
[ Section1 ]
keyS1: valueS1
"""

	CfgSection2 = \
"""[DEFAULT]
key1:value1
key2:value2
[ Section1 ]
keyS1: valueS1
[ Section2 ]
keyS2: valueS2
"""

	def setUp(self):
		self.emptyConfig = SealConfigFile()
		self.config = SealConfigFile()

	def test_empty(self):
		self.assertFalse( self.emptyConfig.has_section("bla") )
		self.assertTrue( self.emptyConfig.has_section("DEFAULT") )
		self.assertEqual(0, len(self.emptyConfig.sections()) )
		self.assertEqual(0, len(self.emptyConfig.items("non existent")))

	def test_default(self):
		self.config.readfp(io.BytesIO(self.__class__.CfgDefaultWithTitle))
		self.assertTrue( self.config.has_section("DEFAULT") )
		self.assertTrue( self.config.has_section("default") )
		self.assertTrue( self.config.has_section("deFAult") )

		items = self.config.items("anything")
		self.assertEqual(2, len(items))
		kv = dict(items)
		self.assertEqual(2, len(kv))
		self.assertEqual("value1", kv.get("key1"))
		self.assertEqual("value2", kv.get("key2"))

	def test_equals(self):
		self.config.readfp( io.BytesIO("key1=value1\n") )
		self.assertEqual("value1", self.config.get("default", "key1"))

	def test_colon(self):
		self.config.readfp( io.BytesIO("key1:value1\n") )
		self.assertEqual("value1", self.config.get("default", "key1"))

	def test_trim_key(self):
		self.config.readfp( io.BytesIO("    key1          :value1\n") )
		self.assertEqual("value1", self.config.get("default", "key1"))

	def test_trim_value(self):
		self.config.readfp( io.BytesIO("key1:     value1       \n") )
		self.assertEqual("value1", self.config.get("default", "key1"))

	def test_section1(self):
		self.config.readfp( io.BytesIO(self.__class__.CfgSection1) )
		self.assertEqual("value1", self.config.get("Section1", "key1"))
		self.assertEqual("valueS1", self.config.get("Section1", "keyS1"))

		kv = dict(self.config.items("Section1"))
		# make sure the iterator goes through all k-v pairs, even the ones inherited from DEFAULT
		self.assertEqual(3, len(kv))
		self.assertEqual("value1", kv.get("key1"))
		self.assertEqual("value2", kv.get("key2"))
		self.assertEqual("valueS1", kv.get("keyS1"))

	def test_section2(self):
		self.config.readfp( io.BytesIO(self.__class__.CfgSection2) )
		self.assertEqual("value1", self.config.get("Section1", "key1"))
		self.assertEqual("valueS1", self.config.get("Section1", "keyS1"))
		self.assertTrue( self.config.get("Section2", "keyS1") is None)
		self.assertEqual("valueS2", self.config.get("Section2", "keyS2"))

	def test_section_that_doesnt_exist(self):
		self.config.readfp( io.BytesIO(self.__class__.CfgSection1) )
		self.assertFalse( self.config.has_section("MySection") )
		self.assertTrue( self.config.get("MySection", "option") is None )

	def test_space_in_section_name(self):
		self.assertRaises(FormatError, self.config.readfp, io.BytesIO("[ Section 1 ]") )

	def test_hash_comment(self):
		self.config.readfp( io.BytesIO(" #key1=value1\nkey2=value2;\n") )
		self.assertTrue( self.config.get("default", "key1") is None)
		self.assertEqual("value2;", self.config.get("default", "key2"))

	def test_semi_colon_comment(self):
		self.config.readfp( io.BytesIO(" ;key1=value1\nkey2=value2;\n") )
		self.assertTrue( self.config.get("default", "key1") is None)
		self.assertEqual("value2;", self.config.get("default", "key2"))

def suite():
	"""Get a suite with all the tests from this module"""
	return unittest.TestLoader().loadTestsFromTestCase(TestSealConfigFile)

if __name__ == '__main__':
	unittest.TextTestRunner(verbosity=2).run(suite())



# Copyright (C) 2011-2012 CRS4.
#
# This file is part of Seal.
#
# Seal is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option)
# any later version.
#
# Seal is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
# for more details.
#
# You should have received a copy of the GNU General Public License along
# with Seal.  If not, see <http://www.gnu.org/licenses/>.

import seal_path
import bl.lib.tools.hadut as hadut
import bl.lib.tools.deprecation_utils as deprecation
from bl.mr.seq.seqal.seqal_config import SeqalConfig, SeqalConfigError

import pydoop.hdfs

import logging
import os
import random
import sys
import tempfile

class SeqalRun(object):

	DefaultReduceTasksPerNode = 6
	LogName = "seqal"
	DefaultLogLevel = 'INFO'

	ConfLogLevel = 'seal.seqal.log.level'
	ConfLogLevel_deprecated = 'bl.seqal.log.level'

	def __init__(self):
		self.parser = SeqalConfig()

		# set default properties
		self.properties = {
			self.ConfLogLevel: self.DefaultLogLevel,
			'hadoop.pipes.java.recordreader': 'true',
			'hadoop.pipes.java.recordwriter': 'true',
			'mapred.create.symlink': 'yes',
			'bl.libhdfs.opts': '-Xmx48m'
		}
		self.hdfs = None
		self.options = None

	def parse_cmd_line(self):
		self.options, self.left_over_args = self.parser.load_config_and_cmd_line()

		# set the job name.  Do it here so the user can override it
		self.properties['mapred.job.name'] = 'seqal_aln_%s' % self.options.output

		# now collect the property values specified in the options and
		# copy them to properties
		for k,v in self.options.properties.iteritems():
			self.properties[k] = v

		# create a logger
		logging.basicConfig()
		self.logger = logging.getLogger(self.__class__.LogName)
		# temporarily set to a high logging level in case we have to print warnings
		# regarding deprecated properties
		self.logger.setLevel(logging.DEBUG)
		# warn for deprecated bl.seqal.log.level property
		if self.properties.has_key(self.ConfLogLevel_deprecated):
			deprecation.deprecation_warning(self.logger, self.ConfLogLevel_deprecated, self.ConfLogLevel)
			if self.properties[self.ConfLogLevel] == self.DefaultLogLevel and \
			    self.properties[self.ConfLogLevel_deprecated] != self.DefaultLogLevel:
				# the deprecated property is different from default, while the new property is not.  Therefore,
				# the user has set the deprecated property to a new value.  We'll use that one.
				self.properties[self.ConfLogLevel] = self.properties[self.ConfLogLevel_deprecated]
				self.logger.warning("Using value %s for property %s (value taken from its deprecated equivalent property %s).",
				    self.properties[self.ConfLogLevel], self.ConfLogLevel, self.ConfLogLevel_deprecated)

		# Set proper logging level
		log_level = getattr(logging, self.properties['seal.seqal.log.level'], None)
		if log_level is None:
			self.logger.setLevel(logging.DEBUG)
			self.logger.warning("Invalid configuration value '%s' for %s.  Check your configuration.", self.ConfLogLevel, self.properties['seal.seqal.log.level'])
			self.logger.warning("Falling back to DEBUG")
			self.logger.warning("Valid values for seal.seqal.log.level are: DEBUG, INFO, WARNING, ERROR, CRITICAL; default: %s", SeqalRun.DefaultLogLevel)
		else:
			self.logger.setLevel(log_level)

		# reference
		self.properties['mapred.cache.archives'] = '%s#reference' % self.options.reference

		# set the number of reduce tasks
		if self.options.align_only:
			n_red_tasks = 0

			if self.options.num_reducers and self.options.num_reducers > 0:
				self.logger.warning("Number of reduce tasks must be 0 when doing --align-only.")
				self.logger.warning("Ignoring request for %d reduce tasks", self.options.num_reducers)
		elif self.options.num_reducers:
			n_red_tasks = self.options.num_reducers
		else:
			n_red_tasks = SeqalRun.DefaultReduceTasksPerNode * hadut.num_nodes()

		self.properties['mapred.reduce.tasks'] = n_red_tasks

	def __write_pipes_script(self, fd):
		ld_path = ":".join( filter(lambda x:x, [seal_path.SealDir, os.environ.get('LD_LIBRARY_PATH', None)]) )
		pypath = os.environ.get('PYTHONPATH', '')
		self.logger.debug("LD_LIBRARY_PATH for tasks: %s", ld_path)
		self.logger.debug("PYTHONPATH for tasks: %s", pypath)

		fd.write("#!/bin/bash\n")
		fd.write('""":"\n')
		fd.write('export LD_LIBRARY_PATH="%s" # Seal dir + LD_LIBRARY_PATH copied from the env where you ran %s\n' % (ld_path, sys.argv[0]))
		fd.write('export PYTHONPATH="%s"\n' % pypath)
		fd.write('exec "%s" -u "$0" "$@"\n' % sys.executable)
		fd.write('":"""\n')
		fd.write('from bl.mr.seq.seqal import run_task\n')
		fd.write('run_task()\n')

	def run(self):
		if self.options is None:
			raise RuntimeError("You must call parse_cmd_line before run")

		if self.logger.isEnabledFor(logging.DEBUG):
			self.logger.debug("Running Seqal")
			self.logger.debug("Properties:\n%s", "\n".join( sorted([ "%s = %s" % (str(k), str(v)) for k,v in self.properties.iteritems() ]) ))
		self.logger.info("Input: %s; Output: %s; reference: %s", self.options.input, self.options.output, self.options.reference)

		try:
			self.hdfs = pydoop.hdfs.hdfs('default', 0)
			self.__validate()

			self.remote_bin_name = tempfile.mktemp(prefix='seqal_bin.', suffix=str(random.random()), dir='')
			try:
				with self.hdfs.open_file(self.remote_bin_name, 'w') as script:
					self.__write_pipes_script(script)

				return hadut.run_pipes(self.remote_bin_name, self.options.input, self.options.output,
					properties=self.properties, args_list=self.left_over_args)
			finally:
				try:
					self.hdfs.delete(self.remote_bin_name) # delete the temporary pipes script from HDFS
					self.logger.debug("pipes script %s deleted", self.remote_bin_name)
				except:
					self.logger.error("Error deleting the temporary pipes script %s from HDFS", self.remote_bin_name)
					## don't re-raise the exception.  We're on our way out
		finally:
			if self.hdfs:
				tmp = self.hdfs
				self.hdfs = None
				tmp.close()
				self.logger.debug("HDFS closed")

	def __validate(self):
		if self.properties['mapred.reduce.tasks'] == 0:
			self.logger.info("Running in alignment-only mode (no rmdup).")

		# validate conditions
		if self.hdfs.exists(self.options.output):
			raise SeqalConfigError("Output directory %s already exists.  Please delete it or specify a different output directory." % self.options.output)
		if not self.hdfs.exists(self.options.reference):
			raise SeqalConfigError("Can't read reference archive %s" % self.options.reference)



# Copyright (C) 2011-2012 CRS4.
#
# This file is part of Seal.
#
# Seal is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Seal is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Seal.  If not, see <http://www.gnu.org/licenses/>.


import unittest
import array

from bl.lib.seq.aligner.io.sam_formatter import SamFormatter
from bl.lib.seq.aligner.mapping import SimpleMapping

class TestSamFormatter(unittest.TestCase):

	def setUp(self):
		self.mapping = SimpleMapping()
		self.f = SamFormatter()

	def test_default(self):
		self.assertEqual("\t".join( map(str, ['', 0, "*", 0, 0, "*", "*", 0, 0, "", "" ])), self.__do_format())

	def test_name(self):
		self.mapping.set_name("my name")
		self.assertEqual("\t".join( map(str, ["my name", 0, "*", 0, 0, "*", "*", 0, 0, "", "" ])), self.__do_format())

	def test_flag(self):
		self.mapping.flag = 163
		self.assertEqual("\t".join( map(str, ['', 163, "*", 0, 0, "*", "*", 0, 0, "", "" ])), self.__do_format())

	def test_reference(self):
		self.mapping.tid = "chr21"
		self.assertEqual("\t".join( map(str, ['', 0, "chr21", 0, 0, "*", "*", 0, 0, "", "" ])), self.__do_format())

	def test_mate_reference(self):
		self.mapping.mtid = "chr21"
		self.assertEqual("\t".join( map(str, ['', 0, "*", 0, 0, "*", "chr21", 0, 0, "", "" ])), self.__do_format())
		self.mapping.mtid = "="
		self.assertEqual("\t".join( map(str, ['', 0, "*", 0, 0, "*", "=", 0, 0, "", "" ])), self.__do_format())

	def test_pos(self):
		self.mapping.pos = 1234
		self.assertEqual("\t".join( map(str, ['', 0, "*", 1234, 0, "*", "*", 0, 0, "", "" ])), self.__do_format())

	def test_mate_pos(self):
		self.mapping.mpos = 1234
		self.assertEqual("\t".join( map(str, ['', 0, "*", 0, 0, "*", "*", 1234, 0, "", "" ])), self.__do_format())

	def test_mapq(self):
		self.mapping.qual = 55
		self.assertEqual("\t".join( map(str, ['', 0, "*", 0, 55, "*", "*", 0, 0, "", "" ])), self.__do_format())

	def test_cigar(self):
		self.mapping.set_cigar([ (100, 'M') ])
		self.assertEqual("\t".join( map(str, ['', 0, "*", 0, 0, "100M", "*", 0, 0, "", "" ])), self.__do_format())
		self.mapping.set_cigar([ (80, 'M'), (20, 'S') ])
		self.assertEqual("\t".join( map(str, ['', 0, "*", 0, 0, "80M20S", "*", 0, 0, "", "" ])), self.__do_format())

	def test_isize(self):
		self.mapping.isize = 273
		self.assertEqual("\t".join( map(str, ['', 0, "*", 0, 0, "*", "*", 0, 273, "", "" ])), self.__do_format())
		self.mapping.isize = -273
		self.assertEqual("\t".join( map(str, ['', 0, "*", 0, 0, "*", "*", 0, -273, "", "" ])), self.__do_format())

	def test_seq(self):
		self.mapping.set_seq_5("AGCTNN")
		self.assertEqual("\t".join( map(str, ['', 0, "*", 0, 0, "*", "*", 0, 0, "AGCTNN", "" ])), self.__do_format())

	def test_qual(self):
		bq = array.array('B', [22, 33, 45])
		self.mapping.set_base_qualities(bq)
		self.assertEqual("\t".join( map(str, ['', 0, "*", 0, 0, "*", "*", 0, 0, "", "7BN" ])), self.__do_format())

	def test_tags(self):
		self.mapping.add_tag( ("XC", 'i', 18) )
		self.mapping.add_tag( ("HI", 'Z', "llo") )
		self.assertEqual("\t".join( map(str, ['', 0, "*", 0, 0, "*", "*", 0, 0, "", "", "XC:i:18", "HI:Z:llo" ])), self.__do_format())

	def __do_format(self):
		return self.f.format(self.mapping)

def suite():
	"""Get a suite with all the tests from this module"""
	return unittest.TestLoader().loadTestsFromTestCase(TestSamFormatter)

if __name__ == '__main__':
	unittest.TextTestRunner(verbosity=2).run(suite())



# Copyright (C) 2011-2012 CRS4.
#
# This file is part of Seal.
#
# Seal is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Seal is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Seal.  If not, see <http://www.gnu.org/licenses/>.

from bl.mr.lib.hit_processor_chain_link import HitProcessorChainLink

class FilterLink(HitProcessorChainLink):
	def __init__(self, monitor, next_link = None):
		super(type(self), self).__init__(next_link)
		self.min_hit_quality = 1
		self.remove_unmapped = True # if true, all unmapped are removed regardless of hit quality
		self.event_monitor = monitor

	def __remove_i(self, pair, i):
		pair[i] = None
		other_hit = pair[i^1]
		if other_hit:
			other_hit.remove_mate()
		return pair

	def process(self, pair):
		if len(pair) != 2:
			raise ValueError("pair length != 2 (it's %d)" % len(pair))
		pair = list(pair) # tuples can't be modified
		for i in 0,1:
			if self.remove_unmapped and pair[i].is_unmapped():
				pair = self.__remove_i(pair, i)
				self.event_monitor.count("reads filtered: unmapped")
			elif pair[i].qual < self.min_hit_quality:
				pair = self.__remove_i(pair, i)
				self.event_monitor.count("reads filtered: low quality")

		if self.next_link and any(pair):
			self.next_link.process(tuple(pair)) # forward pair to next element in chain



# Copyright (C) 2011-2012 CRS4.
#
# This file is part of Seal.
#
# Seal is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Seal is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Seal.  If not, see <http://www.gnu.org/licenses/>.


from bl.mr.lib.hit_processor_chain_link import HitProcessorChainLink
from bl.lib.seq.aligner.io.sam_formatter import SamFormatter

class EmitSamLink(HitProcessorChainLink):
	def __init__(self, context, event_monitor, next_link = None):
		super(type(self), self).__init__(next_link)
		self.ctx = context
		self.output_formatter = SamFormatter(strip_pe_tag=True)
		self.event_monitor = event_monitor

	def process(self, pair):
		for hit in pair:
			if hit:
				k, v = self.output_formatter.format(hit).split("\t", 1)
				self.ctx.emit(str(k), str(v))
				self.event_monitor.count("emitted sam records", 1)

		super(type(self), self).process(pair) # forward pair to next element in chain



# Copyright (C) 2011-2012 CRS4.
#
# This file is part of Seal.
#
# Seal is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Seal is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Seal.  If not, see <http://www.gnu.org/licenses/>.

import sys, os, itertools
import numpy as np
import Bio.SeqIO
import bl.lib.seq.aligner.bwa as bwa
import testing_data as d
import testing_utilities as u


BWA_EXE = "../../../../../../bl/lib/seq/aligner/bwa/libsrc-0.5.7/bwa"


def generate_ref(root):
  f = open(root, 'w')
  f.write(d.refseq)
  f.close()
  os.system("%s index %s" % (BWA_EXE, root))


def read_seq_pairs(seq_reader, N):
  pairs = []
  i = 0
  while i < N:
    p = tuple(itertools.islice(seq_reader, 2))
    if len(p) == 0:
      break
    if len(p) == 1:
      raise ValueError("Uneven number of sequences in input stream!")
    pairs.append(p)
    i += 1
  return pairs

def print_sams(gopt, bns, n_seq, bwsa):
  for i in range(n_seq):
    print bwa.analyze_hit(gopt, bns, bwsa[0][i], bwsa[1][i])
    print bwa.analyze_hit(gopt, bns, bwsa[1][i], bwsa[0][i])

def process_sequences(bwts, bns, pacseq, seq_reader, N, analyze_seqs=None):
  gopt = bwa.gap_init_opt()
  popt = bwa.pe_init_opt()

  ii      = bwa.isize_info_t()
  last_ii = bwa.isize_info_t()
  last_ii.avg = -1.0

  while 1:
    pairs = read_seq_pairs(seq_reader, N)
    seq_pairs_read = len(pairs)
    if seq_pairs_read == 0:
      break
    bwsa = bwa.build_bws_array(pairs)
    bwa.cal_sa_reg_gap(0, bwts, seq_pairs_read, bwsa[0], gopt)
    bwa.cal_sa_reg_gap(0, bwts, seq_pairs_read, bwsa[1], gopt)
    cnt_chg = bwa.cal_pac_pos_pe(bwts, seq_pairs_read, bwsa,
                                 ii, popt, gopt, last_ii)
    sys.stderr.write('ii: %r\n' %[ii.avg, ii.std, ii.low, ii.high])
    bwa.paired_sw(bns, pacseq, seq_pairs_read, bwsa, popt, ii)
    bwa.refine_gapped(bns, seq_pairs_read, bwsa[0], pacseq)
    bwa.refine_gapped(bns, seq_pairs_read, bwsa[1], pacseq)
    analyze_seqs(gopt[0], bns, seq_pairs_read, bwsa)
    bwa.free_seq(N, bwsa[0])
    bwa.free_seq(N, bwsa[1])


def main(argv):
  try:
    hit_fn = argv[1]
  except IndexError:
    hit_fn = 'hit_pairs.fastq'
  N = 10
  root_name = 'refseq'
  generate_ref(root_name)
  bwts = bwa.restore_index(root_name)
  bnsp, pacseq = bwa.restore_reference(root_name)
  sys.stderr.write(str(bnsp[0]))
  seq_reader = Bio.SeqIO.parse(open(hit_fn), 'fastq-illumina')
  process_sequences(bwts, bnsp, pacseq, seq_reader, N, print_sams)
  u.clean_up_aux_files(root_name)

if __name__ == "__main__":
  main(sys.argv)



# Copyright (C) 2011-2012 CRS4.
#
# This file is part of Seal.
#
# Seal is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Seal is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Seal.  If not, see <http://www.gnu.org/licenses/>.

import sys, itertools as it
import Bio.SeqIO

import bl.lib.seq.aligner.bwa as bwa
from bl.lib.seq.aligner.bwa.bwa_iterator import BWAIterator


def main(argv):
  try:
    refseq_fname = argv[1]
    read_fname = argv[2]
    mate_fname = argv[3]
  except IndexError:
    sys.exit("Usage: %s REFSEQ_FN READ_FN MATE_FN" % sys.argv[0])

  seq_list_len = 10000
  max_isize = pairing_batch_size = 10000
  gopt, popt = bwa.gap_init_opt(), bwa.pe_init_opt()

  read_flow = Bio.SeqIO.parse(open(read_fname), 'fastq-illumina')
  mate_flow = Bio.SeqIO.parse(open(mate_fname), 'fastq-illumina')
  pairs_flow = it.izip(read_flow, mate_flow)
  pairs = list(it.islice(pairs_flow, 0, seq_list_len))
  bwts = bwa.restore_index(refseq_fname)
  bnsp, pacseq = bwa.restore_reference(refseq_fname)

  l = len(pairs)
  bwsa = bwa.build_bws_array(pairs)

  bwa_iterator = BWAIterator(refseq_fname, gopt, popt, max_isize,
                             pairing_batch_size)
  pairs = [p for p in bwa_iterator.analyze(bwsa, l)]
  print "READ POS GAPO GAPE MM STRAND SCORE CIGAR"
  for read, mate in pairs:
    if read.n_multi > 0:
      print
      multi_list = [m for m in read.itermulti()]
      for m in multi_list:
        print read.get_name(), m.pos, m.n_gapo, m.n_gape, m.n_mm, m.strand, \
               m.score, m.get_cigar(read.len)

  for j in 0, 1:
    bwa.free_seq(l, bwsa[j])
  bwa.bns_destroy(bwa_iterator.bnsp)


if __name__ == "__main__":
  main(sys.argv)



# Copyright (C) 2011-2012 CRS4.
#
# This file is part of Seal.
#
# Seal is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Seal is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Seal.  If not, see <http://www.gnu.org/licenses/>.

import os
import glob

import bwa_core as bwa
from bwa_iterator import BWAIterator
from bl.lib.tools.event_monitor import QuietMonitor

BWA_INDEX_MANDATORY_EXT = set(["amb", "ann", "bwt", "pac", "rbwt", "rpac"])
BWA_INDEX_MMAP_EXT = set(["rsax", "sax"])
BWA_INDEX_NORM_EXT = set(["rsa", "sa"])
BWA_INDEX_EXT = BWA_INDEX_MANDATORY_EXT | BWA_INDEX_MMAP_EXT | BWA_INDEX_NORM_EXT

class BwaAligner(object):
  """
  Object oriented interface to perform bwa aln + bwa sampe
  using libbwa and the bwa module.
  """
  def __init__(self):
    self.event_monitor = QuietMonitor()
    self.logger = QuietMonitor()
    self.__gopt = bwa.gap_init_opt()
    self.__popt = bwa.pe_init_opt()
    self.__batch = []

    self.qformat = "fastq-illumina"
    self.max_isize = 1000
    self.nthreads = 1
    self.trim_qual = 0
    self.mmap_enabled = False
    # .reference is the path to the indexed BWA reference to use.  Include the
    # "root name" of the reference files--i.e. the names without the file extensions.
    # The trailing dot, if included, will be removed.
    # e.g. my_references/hg_18.bwt
    #      my_references/hg_18.rsax
    #      my_references/hg_18.sax   => "my_references/hg_18"
    #      my_references/hg_18.pac
    #      ...
    self.reference = None
    self.hit_visitor = None
    self.__iterator = None

  def load_pair_record(self, record):
    """
    Append a tuple of the format (id, seq1, qual1, seq2, qual2) to this aligner's work batch.
    """
    self.__batch.append(record)
    return len(self.__batch)

  def get_batch_size(self):
    return len(self.__batch)

  def clear_batch(self):
    self.__batch = []

  def run_alignment(self):
    if not self.reference:
      raise ValueError("You must set the reference path before calling run_alignment")
    if not self.hit_visitor:
      raise ValueError("You must set the hit_visitor before calling run_alignment (else you'll lose the alignment results)")

    self.__check_reference()

    # update __gopt to reflect any changes to our public attributes.
    self.__gopt.contents.n_threads = self.nthreads
    self.event_monitor.log_debug("BWA using %d threads", self.__gopt.contents.n_threads)

    if self.__iterator is None:
      self.__iterator = BWAIterator(self.reference, self.__gopt, self.__popt,
                                    self.max_isize, len(self.__batch),
                                    visitor=self.event_monitor,
                                    mmap_enabled=self.mmap_enabled)

    status = "aligning against %s" % os.path.basename(self.reference)
    self.event_monitor.new_status(status)
    self.event_monitor.log_debug(status)

    self.event_monitor.count("reads processed", 2*len(self.__batch))

    with self.event_monitor.time_block("build_bwsa"):
      # we need to rebuild sequences each time, bwa functions modify them.
      bwsa = bwa.build_bws_array(self.__batch, qtype=self.qformat, trim_qual=self.trim_qual)

    self.__count_bases(bwsa, len(self.__batch))

    # bwa_iterator.analyze_pairs performs the alignment, then lets you iterate
    # over the results
    with self.event_monitor.time_block("analyze_pairs (cal_+sw+refgap+process)"):
      # The python speed optimization tips suggest removing non-local variable look-ups
      # from loops.
      # http://wiki.python.org/moin/PythonSpeed/PerformanceTips
      h_proc = self.hit_visitor.process
      for hit1, hit2 in self.__iterator.analyze_pairs(bwsa, len(self.__batch)):
        h_proc( (hit1, hit2) )

    with self.event_monitor.time_block("destroy_sequences"):
      for j in 0, 1:
        bwa.free_seq(len(self.__batch), bwsa[j])

  def release_resources(self):
    if self.__iterator:
      self.__iterator.unload_reference()

  def __count_bases(self, bwsa, bwsa_size):
    self.event_monitor.start("count_bases")
    for r in 0,1:
      for i in xrange(bwsa_size):
        seq = bwsa[r][i]
        self.event_monitor.count("total bases", seq.full_len)
        self.event_monitor.count("trimmed bases", seq.full_len - seq.clip_len)
    self.event_monitor.stop("count_bases")


  def __check_reference(self):
    """
    Checks that the indexed referenced at path self.reference includes
    files with all the expected extensions.
    Raises ValueError if something is missing.
    """
    if self.reference[-1] == '.':
      self.reference = self.reference[0:-1] # remove the trailing '.', if any
    roots = set([ os.path.splitext(path)[0] for path in glob.iglob(self.reference + ".*") ])
    if len(roots) != 1:
      raise ValueError("BWA index files must differ only by extension.  Found roots: %s" % (roots,))
    extensions = set([ os.path.splitext(path)[1].lstrip('.') for path in glob.iglob(self.reference + ".*") ])

    missing = BWA_INDEX_MANDATORY_EXT - extensions
    if missing:
      raise ValueError("Missing BWA index file types: %s" % (tuple(missing),))
    if self.mmap_enabled and (BWA_INDEX_MMAP_EXT - extensions):
      raise ValueError("Missing BWA mmap index files: %s" % (tuple(BWA_INDEX_MMAP_EXT - extensions),))
    elif not self.mmap_enabled and (BWA_INDEX_NORM_EXT - extensions):
      raise ValueError("Missing BWA index files: %s" % (tuple(BWA_INDEX_NORM_EXT - extensions),))



# Copyright (C) 2011-2012 CRS4.
#
# This file is part of Seal.
#
# Seal is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Seal is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Seal.  If not, see <http://www.gnu.org/licenses/>.


import unittest
from bl.mr.lib.hit_processor_chain_link import HitProcessorChainLink

class TestHitProcessorChainLink(unittest.TestCase):
	def setUp(self):
		self.h1 = HitProcessorChainLink()
		self.h2 = HitProcessorChainLink()

	def test_default_constructor(self):
		self.assertTrue(self.h1.next_link is None)

	def test_constructor(self):
		h = HitProcessorChainLink(self.h1)
		self.assertEqual(self.h1, h.next_link)

	def test_set_next(self):
		self.assertTrue(self.h1.next_link is None)
		retval = self.h1.set_next(self.h2)
		self.assertEqual(self.h2, retval)
		self.assertEqual(self.h2, self.h1.next_link)

	def test_process_no_next(self):
		self.h1.process("test") # shouldn't raise

	def test_process_next(self):
		class Receiver(object):
			def process(self, pair):
				self.received = pair

		receiver = Receiver()
		self.h1.set_next(receiver)
		self.h1.process("test")
		self.assertEqual(receiver.received, "test")

def suite():
	return unittest.TestLoader().loadTestsFromTestCase(TestHitProcessorChainLink)

if __name__ == '__main__':
	unittest.TextTestRunner(verbosity=2).run(suite())



# Copyright (C) 2011-2012 CRS4.
#
# This file is part of Seal.
#
# Seal is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Seal is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Seal.  If not, see <http://www.gnu.org/licenses/>.

import unittest

from bl.mr.lib.hit_processor_chain_link import HitProcessorChainLink
from bl.mr.lib.filter_link import FilterLink
from bl.lib.seq.aligner.mapping import SimpleMapping
from bl.lib.tools.standard_monitor import StandardMonitor
from bl.mr.test_utils import SavingLogger

class TestFilterLink(unittest.TestCase):

	# mini object to let us peek at what the filter forwards to the next link
	class Receiver(HitProcessorChainLink):
		def __init__(self, *args):
			super(type(self), self).__init__(*args)
			self.received = None

		def process(self, pair):
			self.received = pair

	def setUp(self):
		self.monitor = StandardMonitor(SavingLogger())
		self.filter = FilterLink(self.monitor)
		self.receiver = self.filter.set_next(type(self).Receiver())
		# create two mappings, m1, m2.  We put them in self.pair
		# m1 has:
		#   name = first
		# 	tid = tid1
		# m2 has:
		#   name = second
		#   tid = tid2
		self.pair = [ SimpleMapping(), SimpleMapping() ]
		self.m1, self.m2 = self.pair
		self.m1.set_name("first") ; self.m2.set_name("second")
		self.m1.tid = "tid1" ; self.m2.tid = "tid2"
		self.m1.qual = 50 ; self.m2.qual = 30

	def test_constructor_link(self):
		h = FilterLink(self.monitor)
		self.assertTrue(h.next_link is None)
		other = HitProcessorChainLink()
		h = FilterLink(self.monitor, other)
		self.assertEqual(other, h.next_link)

	def test_filter_none(self):
		self.filter.process(self.pair)
		self.assertFalse(self.receiver.received is None)
		self.assertEqual(self.m1.get_name(), self.receiver.received[0].get_name())
		self.assertEqual(self.m2.get_name(), self.receiver.received[1].get_name())
		# ensure there are no counters (i.e. nothing was filtered)
		self.assertFalse( [ c for c in self.monitor.each_counter() ] )

	def test_filter_one(self):
		self.filter.min_hit_quality = self.m2.qual + 1
		self.filter.process(self.pair)
		self.assertFalse(self.receiver.received is None)
		self.assertTrue(self.receiver.received[1] is None)
		self.assertEqual(self.m1.get_name(), self.receiver.received[0].get_name())
		counter_list = [ c for c in self.monitor.each_counter() ]
		self.assertTrue(len(counter_list) == 1)
		name, value = counter_list[0]
		self.assertEqual("reads filtered: low quality", name)
		self.assertEqual(1, value)


	def test_filter_two(self):
		self.filter.min_hit_quality = self.m1.qual + 1
		self.filter.process(self.pair)
		self.assertTrue(self.receiver.received is None)
		counter_list = [ c for c in self.monitor.each_counter() ]
		self.assertTrue(len(counter_list) == 1)
		name, value = counter_list[0]
		self.assertEqual("reads filtered: low quality", name)
		self.assertEqual(2, value)

	def test_without_next_link(self):
		h = FilterLink(self.monitor)
		h.process(self.pair) # shouldn't raise

	def test_filter_unmapped_1(self):
		self.m1.set_mapped(False)
		self.m1.qual = 0
		self.filter.process(self.pair)
		self.assertTrue(self.receiver.received[0] is None)
		self.assertFalse(self.receiver.received[1] is None)
		counter_list = [ c for c in self.monitor.each_counter() ]
		self.assertTrue(len(counter_list) == 1)
		name, value = counter_list[0]
		self.assertEqual("reads filtered: unmapped", name)
		self.assertEqual(1, value)


def suite():
	return unittest.TestLoader().loadTestsFromTestCase(TestFilterLink)

if __name__ == '__main__':
	unittest.TextTestRunner(verbosity=2).run(suite())



# Copyright (C) 2011-2012 CRS4.
#
# This file is part of Seal.
#
# Seal is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Seal is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Seal.  If not, see <http://www.gnu.org/licenses/>.

import unittest
import re

from bl.mr.lib.emit_sam_link import EmitSamLink
from bl.mr.lib.hit_processor_chain_link import HitProcessorChainLink
from bl.mr.lib.hadoop_event_monitor import HadoopEventMonitor
from bl.mr.test_utils import map_context, SavingLogger
from bl.lib.seq.aligner.mapping import *

class TestEmitSamLink(unittest.TestCase):
	def setUp(self):
		self.map_ctx = map_context(None, None)
		self.count_group = "Test"
		self.logger = SavingLogger()
		self.monitor = HadoopEventMonitor(self.count_group, self.logger, self.map_ctx)
		self.emitter = EmitSamLink(self.map_ctx, self.monitor)
		# create two mappings, m1, m2.  We put them in self.pair
		# m1 has:
		#   name = first
		# 	tid = tid1
		# m2 has:
		#   name = second
		#   tid = tid2
		self.pair = [ SimpleMapping(), SimpleMapping() ]
		self.m1, self.m2 = self.pair
		self.m1.set_name("first")
		self.m1.tid = "tid1"
		self.m2.set_name("second")
		self.m2.tid = "tid2"

	def test_constructor_link(self):
		h = EmitSamLink(self.map_ctx, self.monitor)
		self.assertTrue(h.next_link is None)
		other = HitProcessorChainLink()
		h = EmitSamLink(self.map_ctx, self.monitor, other)
		self.assertEqual(other, h.next_link)

	def test_process(self):
		self.emitter.process(self.pair)
		self.assertEqual(["first", "second"], sorted(self.map_ctx.emitted.keys()))
		self.assertEqual(1, len(self.map_ctx.emitted["first"]))
		self.assertTrue(re.search("tid1", self.map_ctx.emitted["first"][0]))
		self.assertEqual(1, len(self.map_ctx.emitted["second"]))
		self.assertTrue(re.search("tid2", self.map_ctx.emitted["second"][0]))

	def test_emitted_type(self):
		self.emitter.process(self.pair)
		for k in self.map_ctx.emitted.keys():
			self.assertTrue(isinstance(k, str))
		for v in [ item for ary in self.map_ctx.emitted.values() for item in ary ]:
			self.assertTrue(isinstance(v, str))

	def test_first_null(self):
		self.pair[0] = None
		self.emitter.process(self.pair)
		self.assertEqual(["second"], self.map_ctx.emitted.keys())
		self.assertEqual(1, len(self.map_ctx.emitted["second"]))
		self.assertTrue(re.search("tid2", self.map_ctx.emitted["second"][0]))

	def test_second_null(self):
		self.pair[1] = None
		self.emitter.process(self.pair)
		self.assertEqual(["first"], self.map_ctx.emitted.keys())
		self.assertEqual(1, len(self.map_ctx.emitted["first"]))
		self.assertTrue(re.search("tid1", self.map_ctx.emitted["first"][0]))

	def test_forward_pair(self):
		class Receiver(object):
			def process(self, pair):
				self.received = pair

		receiver = Receiver()
		self.emitter.set_next(receiver)
		self.emitter.process(self.pair)
		self.assertEqual(self.pair, receiver.received)

def suite():
	return unittest.TestLoader().loadTestsFromTestCase(TestEmitSamLink)

if __name__ == '__main__':
	unittest.TextTestRunner(verbosity=2).run(suite())



#!/usr/bin/env python

# Copyright (C) 2011-2012 CRS4.
#
# This file is part of Seal.
#
# Seal is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Seal is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Seal.  If not, see <http://www.gnu.org/licenses/>.


import sys, os, unittest, imp

D = os.path.dirname(__file__)
BuildDir = os.path.realpath(os.path.join(D, "..", "build"))
sys.path.insert(0, BuildDir)

try:
	import bl.lib.tools.standard_monitor
except ImportError:
	print >>sys.stderr, "Can't import bl module.  Did you build Seal? (call 'make' in the Seal root directory). If you installed Seal somewhere, please export PYTHONPATH to include the installation path."
	sys.exit(1)


TEST_MODULES = [os.path.join(D, m) for m in [

  "bl/lib/seq/aligner/bwa/test_bwa_mapping.py",
  "bl/lib/seq/aligner/bwa/test_core.py",
  "bl/lib/seq/aligner/bwa/test_bwa_aligner.py",  # currently broken on some 32-bit systems, see #62
  "bl/lib/seq/aligner/io/test_protobuf_mapping.py",
  "bl/lib/seq/aligner/io/test_sam_formatter.py",
  "bl/lib/seq/aligner/test_mapping.py",
  "bl/lib/tools/test_seal_config_file.py",
  "bl/mr/lib/test_emit_sam_link.py",
  "bl/mr/lib/test_filter_link.py",
  "bl/mr/lib/test_hadoop_event_monitor.py",
  "bl/mr/lib/test_hit_processor_chain_link.py",
  "bl/mr/seq/seqal/test_mark_duplicates_emitter.py",
  "bl/mr/seq/seqal/test_reducer.py",

## add new unit test modules here.  They must provide a suite() method
## that returns a unittest.TestSuite instance.  Paths are relative to
## this module's directory

  ]]


class UnitTestRunner(object):

  def __init__(self):
    self.autotest_list = TEST_MODULES

  @staticmethod
  def __load_suite(module_path):
    module_name = os.path.splitext(os.path.basename(module_path))[0]
    ## so that test modules can import other modules in their own
    ## directories, we directly modify sys.path
    sys.path.append(os.path.dirname(module_path))
    fp, pathname, description = imp.find_module(module_name)
    try:
      module = imp.load_module(module_name, fp, pathname, description)
      del sys.path[-1]  # clean up to avoid conflicts
      return module.suite()
    finally:
      fp.close()

  def run(self):
    suites = map(UnitTestRunner.__load_suite, self.autotest_list)
    unittest.TextTestRunner(verbosity=2).run(unittest.TestSuite(tuple(suites)))


if __name__ == '__main__':
  UnitTestRunner().run()



# Copyright (C) 2011-2012 CRS4.
#
# This file is part of Seal.
#
# Seal is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Seal is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Seal.  If not, see <http://www.gnu.org/licenses/>.

from bl.lib.tools.event_monitor import EventMonitor
from bl.lib import TIME_TICK

import time
import logging
logging.basicConfig(level=logging.DEBUG)

class HadoopEventMonitor(EventMonitor):
	"""
	Event monitor implementation to log information through the
	Hadoop context.
	"""
	def __init__(self, event_class, logger, ctx):
		"""
		Parameters:
		event_class:  the category in which all counters will be created.
		logger:  logger object to use
		ctx:  Hadoop context
		"""
		self.event_class = event_class
		self.logger = logger
		self.ctx = ctx
		self.__start_times = {}
		self.__counters = {}

	def __get_counter(self, name, timing=False):
		if not self.__counters.has_key(name):
			if timing:
				self.add_counter(name, "TIME_" + name.upper())
			else:
				self.add_counter(name, name.upper())
		return self.__counters[name]

	def __get_timing_counter(self, name):
		return self.__get_counter(name, timing=True)

	def start(self, s):
		self.__start_times[s] = time.time()

	def stop(self, s):
		delta = time.time() - self.__start_times[s]
		self.ctx.incrementCounter(self.__get_timing_counter(s), int(delta/TIME_TICK))
		status = "done with %s (%.3f s)" % (s, delta)
		self.ctx.setStatus(status)
		self.log_debug(status)

	def stop_batch(self, s, offset, n):
		delta = time.time() - self.__start_times[s]
		self.ctx.incrementCounter(self.__get_timing_counter(s), int(delta/TIME_TICK))
		status = "done with %s (offset=%d, n=%d) (%.3f s)" % (s, offset, n, delta)
		self.ctx.setStatus(status)
		self.log_debug(status)

	def count(self, name, value=1):
		self.ctx.incrementCounter(self.__get_counter(name), value)

	def has_counter(self, name):
		return self.__counters.has_key(name)

	def add_counter(self, name, display_name=None):
		if self.__counters.has_key(name):
			raise ValueError("counter '%s' is already defined" % name)
		self.__counters[name] = self.ctx.getCounter(self.event_class, display_name if display_name else name.upper())

	#################################
	# logging methods
	#################################
	def new_status(self, status_msg):
		self.ctx.setStatus(status_msg)
		self.log_debug(status_msg)

	def log_debug(self, *args):
		self.logger.debug(*args)

	def log_info(self, *args):
		self.logger.info(*args)

	def log_warning(self, *args):
		self.logger.warning(*args)

	def log_error(self, *args):
		self.logger.error(*args)

	def log_critical(self, *args):
		self.logger.critical(*args)




# Copyright (C) 2011-2012 CRS4.
#
# This file is part of Seal.
#
# Seal is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Seal is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Seal.  If not, see <http://www.gnu.org/licenses/>.

"""Seal: Sequence Alignment on Hadoop.

Seal is a of MapReduce application for biological
sequence alignment. It runs on Hadoop (http://hadoop.apache.org)
through Pydoop (http://pydoop.sourceforge.net), a Python MapReduce
and HDFS API for Hadoop.
"""

import glob
import os
import shutil
import sys
from distutils.core import setup
from distutils.errors import DistutilsSetupError

def get_arg(name):
	arg_start = "%s=" % name
	for i in xrange(len(sys.argv)):
		arg = sys.argv[i]
		if arg.startswith(arg_start):
			value = arg.replace(arg_start, "", 1)
			del sys.argv[i]
			if not value:
				raise RuntimeException("blank value specified for %s" % name)
			return value
	return None

def check_python_version():
	override = ("true" == get_arg("override_version_check"))
	if not override and sys.version_info < (2,6):
		print >>sys.stderr, "Please use a version of Python >= 2.6 (currently using vers. %s)." % ",".join( map(str, sys.version_info))
		print >>sys.stderr, "Specify setup.py override_version_check=true to override this check."
		sys.exit(1)


def get_version():
  vers = get_arg("version")
  if vers is None:
    # else, if no version specified on command line
    version_filename = os.path.join( os.path.dirname(__file__), 'VERSION')
    if os.path.exists(version_filename):
      with open(version_filename) as f:
        vers = f.read().rstrip("\n")
    else:
      from datetime import datetime
      vers = datetime.now().strftime("devel-%Y%m%d")#_%H%M%S")
  return vers

NAME = 'seal'
DESCRIPTION = __doc__.split("\n", 1)[0]
LONG_DESCRIPTION = __doc__
URL = "http://www.crs4.it"
# DOWNLOAD_URL = ""
LICENSE = 'GPL'
CLASSIFIERS = [
  "Programming Language :: Python",
  "License :: OSI Approved :: GNU General Public License (GPL)",
  "Operating System :: POSIX :: Linux",
  "Topic :: Scientific/Engineering :: Bio-Informatics",
  "Intended Audience :: Science/Research",
  ]
PLATFORMS = ["Linux"]
VERSION = get_version()
AUTHOR_INFO = [
  ("Luca Pireddu", "luca.pireddu@crs4.it"),
  ("Simone Leo", "simone.leo@crs4.it"),
  ("Gianluigi Zanetti", "gianluigi.zanetti@crs4.it"),
  ]
MAINTAINER_INFO = [
  ("Luca Pireddu", "luca.pireddu@crs4.it"),
  ]
AUTHOR = ", ".join(t[0] for t in AUTHOR_INFO)
AUTHOR_EMAIL = ", ".join("<%s>" % t[1] for t in AUTHOR_INFO)
MAINTAINER = ", ".join(t[0] for t in MAINTAINER_INFO)
MAINTAINER_EMAIL = ", ".join("<%s>" % t[1] for t in MAINTAINER_INFO)


#-- FIXME: handle internally ------------------------------------------------
from distutils.command.build import build

class seqal_build(build):
  def finalize_options(self):
    build.finalize_options(self)
    global VERSION
    # HACK!  Until we find a better way to pass a parameter 
    # into the build command.
    self.version = VERSION
    self.override_version_check = get_arg("override_version_check") or 'false'
 
  def run(self):
    build.run(self)
    libbwa_dir = "bl/lib/seq/aligner/bwa"
    libbwa_src = os.path.join(libbwa_dir, "libbwa")
    libbwa_dest = os.path.abspath(os.path.join(self.build_purelib, libbwa_dir))
    ret = os.system("BWA_LIBRARY_DIR=%s make -C %s libbwa" %
                    (libbwa_dest, libbwa_src))
    if ret:
      raise DistutilsSetupError("could not make libbwa")
    # protobuf classes
    proto_src = "bl/lib/seq/aligner/io/mapping.proto"
    ret = os.system("protoc %s --python_out=%s" %
                    (proto_src, self.build_purelib))
    if ret:
      raise DistutilsSetupError("could not run protoc")
    ret = os.system('ant -Dversion="%s" -Doverride_version_check="%s"' % (self.version, self.override_version_check))
    if ret:
      raise DistutilsSetupError("Could not build Java components")
    self.package()

  def package(self):
    dest_jar_path = os.path.join(self.build_purelib,'bl', 'seal.jar')
    if os.path.exists(dest_jar_path):
      os.remove(dest_jar_path)
    shutil.move( os.path.join(self.build_base, 'seal.jar'), dest_jar_path )



#############################################################################
# main
#############################################################################

check_python_version()
setup(name=NAME,
      description=DESCRIPTION,
      long_description=LONG_DESCRIPTION,
      url=URL,
##      download_url=DOWNLOAD_URL,
      license=LICENSE,
      classifiers=CLASSIFIERS,
      author=AUTHOR,
      author_email=AUTHOR_EMAIL,
      maintainer=MAINTAINER,
      maintainer_email=MAINTAINER_EMAIL,
      platforms=PLATFORMS,
      version=VERSION,
      packages=['bl',
                'bl.lib',
                'bl.lib.tools',
                'bl.lib.seq',
                'bl.lib.seq.aligner',
                'bl.lib.seq.aligner.bwa',
                'bl.lib.seq.aligner.io',
                'bl.mr',
                'bl.mr.lib',
                'bl.mr.seq',
                'bl.mr.seq.seqal',
                ],
      cmdclass={"build": seqal_build},
      scripts=glob.glob("bin/*"),
      )
